//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#if !defined(WIN32)
  #include "windowsdef.h"
#else
  #include <windows.h>
  #pragma pack(1)
#endif

#include <stdint.h>

#define P1		1
#define P2		2
#define P3		3
#define PK		4

#define SINGLE_CONNECTION		0	// port will be opened for single connection
#define MULTIPLE_CONNECTIONS		1	// port will be opened for multiple connections (until manually closed)

#define MAX_KNOCKS			3
#define NUM_PARANOID_RESENT		2
#define ETHPROTO_IP		0x08



typedef struct __attribute__((packed)) ipheader
{
        unsigned char ip_hl:4, ip_v:4; /* Each member is 4 bits */
        unsigned char ip_tos;
        unsigned short int ip_len;
        unsigned short int ip_id;
        unsigned short int ip_off;
        unsigned char ip_ttl;
        unsigned char ip_p;
        unsigned short int ip_sum;
        uint32_t ip_src;
        uint32_t ip_dst;
} IP_HEADER; /* total ip header length: 20 bytes (=160 bits) */


typedef struct __attribute__((packed)) tcpheader
{
        unsigned short int th_sport;
        unsigned short int th_dport;
        uint32_t th_seq;
        uint32_t th_ack;
        unsigned char th_resv:4, th_headerlen:4;
        unsigned char th_flags;
        unsigned short int th_win;
        unsigned short int th_sum;
        unsigned short int th_urp;
} TCP_HEADER; /* total tcp header length: 20 bytes (=160 bits) */


typedef struct __attribute__((packed)) pseudoheader
{
        uint32_t ip_src;
        uint32_t ip_dst;
        unsigned char resv;
        unsigned char ip_p;
        unsigned short tcp_len;         // length of tcp header + data
} PSEUDO_HEADER;


typedef struct __attribute__((packed)) udpheader
{
	unsigned short int th_sport;
	unsigned short int th_dport;
	unsigned short int th_len;
	unsigned short int th_sum;
} UDP_HEADER;


struct __attribute__((packed)) tcp_pseudoheader
{
        unsigned int ip_src;
        unsigned int ip_dst;
        unsigned char resv;
        unsigned char ip_p;
        unsigned short tcp_len;         // length of tcp header + data
};


#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PSH  0x08
#define TH_ACK  0x10
#define TH_URG  0x20
#define TH_ECN  0x40
#define TH_CWR  0x80


// Format of port sequence request packet.
// This is the first packet the client sents to the server to indicate the
// knock sequence.

typedef struct __attribute__((packed)) _PORTSEQ_REQ
{
	unsigned char IV[16];
	unsigned char MD5[16];
	unsigned char version;	
	unsigned short int packetType;		// P1
	unsigned short int reqType;			// single connection/multiple connections
	uint32_t internelIP;
	unsigned short int ports[MAX_KNOCKS];
	uint32_t ISNs[MAX_KNOCKS];
	uint32_t timeStamp;
	unsigned short int service;
	unsigned short int portLow;
	unsigned short int portHigh;
} PORTSEQ_REQ;


// Format of port request query packet
// This packet is sent by the client after the SYN knocks to query the server for the 
// status of the knocks.

typedef struct __attribute__((packed)) _PORTREQ_QUERY
{
	unsigned char IV[16];
	unsigned char MD5[16];
	unsigned char version;	
	unsigned short int packetType;		// P2
	unsigned short int reqType;			// single connection/multiple connections
	uint32_t internelIP;
	uint32_t timeStamp;
} PORTREQ_QUERY;

// Format of port knock

typedef struct __attribute__((packed)) _PORTKNOCK
{
	unsigned char IV[16];
	unsigned char MD5[16];
	unsigned char version;	
	unsigned short int packetType;		// PK
	unsigned short int reqType;			// single connection/multiple connections
	uint32_t internelIP;
	uint32_t ISN;
	uint32_t timeStamp;
} PORTKNOCK;


// Format of port request query reply packet
// This packet is sent by the server to the client in reply to packet P2.  It tells
// the client the assigned allocated port and the port open time.

typedef struct __attribute__((packed)) _PORTREQ_QUERY_REPLY
{
	unsigned char IV[16];
	unsigned char MD5[16];
	unsigned char version;	
	unsigned short int packetType;		// P3
	unsigned short int reqType;			// single connection/multiple connections
	unsigned short int allocatedPort;
	unsigned short int portOpenTime;	// port open time
	uint32_t timeStamp;
} PORTREQ_QUERY_REPLY;
