//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#if defined(WIN32)
  #include <windows.h>
  #include <winsock2.h>
  #pragma comment(lib,"ws2_32.lib")
  #pragma pack(1)
  #define SIO_RCVALL _WSAIOW(IOC_VENDOR,1)
#else
  #include "windowsdef.h"
  #include <string.h>
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <sys/ioctl.h>
  #include <errno.h>
  #include <unistd.h>
  #include <fcntl.h>
  #include <netinet/in.h>
  #include <limits.h>
  #include <netdb.h>
  #include <arpa/inet.h>
  #include <net/if.h>
  #include <termios.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#include "md5.h"
#include "randomGen.h"
#include "aes.h"
#include "version.h"
#include "aesCBC.h"
#include "user.h"
#include "clientProcessPkt.h"

DWORD gOurIPAddress;
int verbose_flag=0;
unsigned short int clientService=0;
unsigned short int clientPortLow=0;
unsigned short int clientPortHigh=0;


//************************************************************************************
// Generates user's password hash
// MD5(userid + password)
//
// outPasswordHash must have SIZEOF_USERPASSWORD_HASH of storage
//
//************************************************************************************

void compUserPasswordHash(char *userID, char *userPassword, unsigned char *outPasswordHash)
{
	MD5Context ctx;
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *)userID, strlen(userID));
	MD5Update(&ctx, (unsigned char *)userPassword, strlen(userPassword));
						
	MD5Final(outPasswordHash, &ctx);

}


void printUsage(char *exeName)
{
	printf("\nUsage: %s [options] <server IP> <server Port>\n\n", exeName);
	printf("Options are :\n");
	printf("-u username / --user username : set user to username\n");
	printf("-p pw / --password pw : set password to pw\n");
	printf("-s service / --service service : set requested service to service\n");
	printf("-r portRange / --ports portRange : set requested port range (if allowed) to portRange\n");
	printf("\n");
	return;
}


//************************************************************************************
// Gets User ID and Password from Console
//
//************************************************************************************
void getUserIDPassword(char *userID, int userIDSize, char *password, int passwordSize)
{
#if defined(WIN32)
	HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE); 
	HANDLE hStdout = GetStdHandle(STD_OUTPUT_HANDLE); 

	char *userNamePrompt = "User Name: ";
	char *userPasswordPrompt = "Password: ";


    if (strlen(userID)==0)
    {
    	memset(userID, 0, userIDSize);
     	printf("%s", userNamePrompt);
      	fgets(userID, userIDSize, stdin);

       	userID[userIDSize - 1] = 0;
        int i;
        for(i = userIDSize - 1; i >= 0; i--)
        {
       		if(userID[i] == 0x0a || userID[i] == 0x0d)
         			userID[i] = 0;
        }
    }
    
    if (strlen(password)==0)
    {
	   printf("%s", userPasswordPrompt);

	   // disable echoing of password
	   DWORD fdwMode, fdwOldMode;
       GetConsoleMode(hStdin, &fdwOldMode);

       fdwMode = fdwOldMode & ~(ENABLE_ECHO_INPUT); 

       SetConsoleMode(hStdin, fdwMode);

	   memset(password, 0, sizeof(password));	
	   fflush(stdin);
	   fgets(password, passwordSize, stdin);


	   password[passwordSize - 1] = 0;
	   int i;
	   for(i = passwordSize - 1; i >= 0; i--)
	   {
	  	  if(password[i] == 0x0a || password[i] == 0x0d)
			  password[i] = 0;
	   }
	
	   SetConsoleMode(hStdin, fdwOldMode);
	   printf("\n");
    }   
}

#else
	struct termios origFlags, newFlags;
	char *userNamePrompt = "User Name: ";
	char *userPasswordPrompt = "Password: ";

    
	if (strlen(userID)==0)
    {
    memset(userID, 0, userIDSize);
	write(1, userNamePrompt, strlen(userNamePrompt));
	read(0, userID, userIDSize - 1);

	userID[userIDSize - 1] = 0;
	int i;
	for(i = userIDSize - 1; i >= 0; i--)
	{
		if(userID[i] == 0x0a || userID[i] == 0x0d)
			userID[i] = 0;
	}
    }
    if (strlen(password)==0)
    {
	// turn off echo
	tcgetattr(0, &origFlags);

	newFlags = origFlags;
	newFlags.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL | ECHOCTL | ECHOPRT | ECHOKE);

	tcsetattr(0, TCSANOW, &newFlags);

	// clear the buffer
	fd_set fset;
	FD_ZERO(&fset);
	FD_SET(0, &fset);
	struct timeval tv;
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	int status = select(1, &fset, NULL, NULL, &tv);
	
	if(status)
	{
		char buffer[1024];
	        read(0, buffer, sizeof(buffer));
	}
	                                                                                                                 
	
	memset(password, 0, sizeof(password));	
	write(1, userPasswordPrompt, strlen(userPasswordPrompt));	
	read(0, password, passwordSize - 1);


	password[passwordSize - 1] = 0;
	int i;
	for(i = passwordSize - 1; i >= 0; i--)
	{
		if(password[i] == 0x0a || password[i] == 0x0d)
			password[i] = 0;
	}
					
	// turn on echo
	tcsetattr(0, TCSANOW, &origFlags);
    }
}
#endif


int main(int argc, char* argv[])
{
	char userID[SIZEOF_USERID];
	char userPassword[SIZEOF_USERPASSWORD];
	unsigned char userPasswordHash[SIZEOF_USERPASSWORD_HASH];

    memset(userID,0,SIZEOF_USERID);
    memset(userPassword,0,SIZEOF_USERPASSWORD);

	printf("DrKnockC Version %s Copyright (c) 2006 Olivier ROCH VILATO and Chew Keong TAN\n\n", PROG_VERSION);



    int c;
    static struct option long_options[] =
    {
           /* These options set a flag. */
           {"verbose", no_argument,       &verbose_flag, 1},
           {"brief",   no_argument,       &verbose_flag, 0},
               /* These options don't set a flag.
                  We distinguish them by their indices. */
               {"help",no_argument,0,'h'},
               {"user",required_argument,  0, 'u'},
               {"password",  required_argument,0, 'p'},
               {"service",  required_argument, 0, 's'},
               {"ports",  required_argument, 0, 'r'},
               {0, 0, 0, 0}
    };
   /* getopt_long stores the option index here. */
    int option_index = 0;
    while (1)
    {
            c = getopt_long (argc, argv, "hu:p:s:r:",
                            long_options, &option_index);
     
           /* Detect the end of the options. */
           if (c == -1)
             break;
     
           switch (c)
             {
             case 0:
               /* If this option set a flag, do nothing else now. */
               if (long_options[option_index].flag != 0)
                 break;
               printf ("option %s", long_options[option_index].name);
               if (optarg)
                 printf (" with arg %s", optarg);
               printf ("\n");
               break;
     
             case 'h':
               printUsage(argv[0]);
               return 1;
               break;
                  
             case 'u':
               if (optarg)
               { puts ("User set");
                 strncpy(userID,optarg,SIZEOF_USERID);
               }                  
               break;
     
             case 'p':
               if (optarg)
               { puts ("Password set");
                 strncpy(userPassword,optarg,SIZEOF_USERID);
               }                  
               break;
     
             case 's':
               if (optarg)
               { puts("Service set");
                 clientService=atoi(optarg);
               }
               break;
     
             case 'r':
               if (optarg)
               { puts("Port Range set");
                 if (strchr(optarg,'-'))
                 { sscanf(optarg,"%d-%d",&clientPortLow,&clientPortHigh);
                 }
                 else
                 { sscanf(optarg,"%d",&clientPortLow);
                   clientPortHigh=clientPortLow;
                 }
               }
  
               break;
    
     
             case '?':
               /* getopt_long already printed an error message. */
               break;
     
             default:
               abort ();
             }
    }
     
       /* Instead of reporting `--verbose'
          and `--brief' as they are encountered,
          we report the final status resulting from them. */
       if (verbose_flag)
         puts ("verbose flag is set");
     
       /* Print any remaining command line arguments (not options). */
/*       if (optind < argc)
         {
           printf ("non-option ARGV-elements: ");
           while (optind < argc)
             printf ("%s ", argv[optind++]);
           putchar ('\n');
         }
*/

	if((argc-optind) != 2)
	{
		printUsage(argv[0]);
		return 1;
	}

	#if defined(WIN32)
	WSADATA wsd;
	#else
	if(!createRawSendSocket())
		return 1;

	// drop priv
	setreuid(getuid(), getuid());
	setregid(getgid(), getgid());
	#endif	

	char *destIP = argv[optind++];
	WORD destPort = atoi(argv[optind++]);

	if(!initRand())
	{
		printf("Error initializing random generator!\n");
		return 1;
	}
	#if defined(WIN32)
	if (WSAStartup(MAKEWORD(2,2), &wsd) != 0)
	{
		printf("WSAStartup() failed: %d\n", GetLastError());
		return -1;
	}
	
	// get our first IP address
	char name[256];
	gethostname(name, sizeof(name));

	struct hostent *pHostent;
	pHostent = gethostbyname(name);
	gOurIPAddress = (*((struct in_addr *)pHostent->h_addr_list[0])).S_un.S_addr;

	#else
	// get our IP address
	getOutInterfaceIP(inet_addr(destIP), &gOurIPAddress);
	#endif

	getUserIDPassword(userID, sizeof(userID), userPassword, sizeof(userPassword));
	if(strlen(userID) == 0 || strlen(userPassword) == 0)
	{
		printf("\nInvalid UserID or Password.\n");
		return 1;
	}
	printf("\n\n");

	compUserPasswordHash(userID, userPassword, userPasswordHash);
	memset(userPassword, 0, sizeof(userPassword));		// clear out

	PORTSEQ_REQ pq, netorder_pq;	
	generateKnockSequence(&pq, &netorder_pq, userID, userPasswordHash);	

	// send encrypt knock seq
	if(doUDPPortKnockReq(destIP, destPort, &netorder_pq, userID, userPasswordHash))
	{


//		if(doSYNPortKnock(destIP, &pq))		// send knocks
		if(doUDPPortKnock(destIP, &pq,userID,userPasswordHash))		// send knocks
		{			
			PORTREQ_QUERY prq, netorder_prq;

			generateKnockResultQuery(&prq, &netorder_prq, userID, userPasswordHash, pq.timeStamp);

			WORD allocPort;
			WORD portOpenTime;
			WORD reqType;

			// send query and wait for reply
			if(doUDPPortKnockQuery(destIP, destPort, &prq, &netorder_prq, userID, userPasswordHash, 
				                   &allocPort, &portOpenTime, &reqType))
			{
				memset(userID, 0, sizeof(userID));
				memset(userPasswordHash, 0, sizeof(userPasswordHash));

				printf("\nDoor is open at %s Port %d for %d seconds.\n", 
					   destIP, allocPort, portOpenTime);

/*				for(int timeCount = portOpenTime; timeCount > 0; timeCount--)
				{
					printf("%d, ", timeCount);
					#if defined(WIN32)
					Sleep(1000);
					#else
					fflush(stdout);
					sleep(1);
					#endif
				}
				printf("\n");
*/			}
			else
			{
				printf("\nSorry, the door is still closed.\n");
			}
		}
	}

	#if defined(WIN32)
	WSACleanup();
	#endif		
	return 0;
}
