//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <string.h>
#include "md5.h"
#include "aes.h"
#include "userRecord.h"


//************************************************************************************
// Generates the 128-bit AES encryption key userID and user's password hash.
//
// outKey must have 16 bytes of storage
//************************************************************************************

void genEncKey(char *userID, unsigned char *userPasswordHash, unsigned char *outKey)
{
	MD5Context ctx;
	MD5Init(&ctx);

	MD5Update(&ctx, (unsigned char *)userPasswordHash, SIZEOF_USERPASSWORD_HASH);
	MD5Update(&ctx, (unsigned char *)userID, strlen(userID));

	MD5Final(outKey, &ctx);
}


//************************************************************************************
// AES Decrypt in CBC mode
//
// IV = 0
//************************************************************************************

void aesDecrypt(aes_ctx *ctx, unsigned char *data, int dataSize)
{
	unsigned char iv[16];	

	memset(iv, 0, sizeof(iv));
	

	for(int i = 0; i < dataSize; i += 16)
	{		
		unsigned char outData[16];

		aes_dec_blk(&data[i], outData, ctx);

		for(int j = 0; j < 16; j++)
		{
			outData[j] ^= iv[j];
		}

		memcpy(iv, &data[i], 16);

		memcpy(&data[i], outData, sizeof(outData));
	}
}


//************************************************************************************
// AES Encrypt in CBC mode
//
// IV = 0
//************************************************************************************

void aesEncrypt(aes_ctx *ctx, unsigned char *data, int dataSize)
{	
	unsigned char iv[16];	

	memset(iv, 0, sizeof(iv));

	for(int i = 0; i < dataSize; i += 16)
	{		
		unsigned char outData[16];
		unsigned char *tempData;

		tempData = &data[i];

		for(int j = 0; j < 16; j++)
			tempData[j] ^= iv[j];

		aes_enc_blk(tempData, outData, ctx);

		memcpy(&data[i], outData, sizeof(outData));

		memcpy(iv, outData, 16);		
	}
}
