//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************


#include <stdio.h>
#include <time.h>
#if defined(WIN32)
  #include <windows.h>
  #include <wincrypt.h>
  HCRYPTPROV		ghProvider = 0;
#else
  #include <stdlib.h>
  FILE *fpRandom = NULL;
#endif

#include "randomGen.h"


//************************************************************************************
// Initializes the crypto context
//
//************************************************************************************

BOOL initRand()
{
	srand(time(NULL));

    #if defined(WIN32)
       	return CryptAcquireContext(&ghProvider, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT);
    #else
  	  fpRandom = fopen("/dev/urandom", "rb");
	  if(fpRandom)
		return TRUE;
	  else
		return FALSE;
    #endif
}

//************************************************************************************
// Returns a DWORD random number
//
//************************************************************************************

DWORD getRandDWORD(void)
{
	DWORD result;

    #if defined(WIN32)
	  /* The CryptoAPI returns sizeof(buf) bytes of randomness */
  	  if(ghProvider)
	  {
		
		if (CryptGenRandom(ghProvider, sizeof(DWORD), (BYTE *)&result))
		{
			return result;
		}
	  }
    #else
	  if(fpRandom)
   	  {
		if(fread(&result, 1, sizeof(DWORD), fpRandom))
		{
			return result;
		}
	  }
    #endif
	
	return (((DWORD)rand()) << 16) | (DWORD)rand();
}

//************************************************************************************
// Returns a WORD random number
//
//************************************************************************************

WORD getRandWORD(void)
{
	WORD result;

    #if defined(WIN32)
	  /* The CryptoAPI returns sizeof(buf) bytes of randomness */
  	  if(ghProvider)
	  {
		
		if (CryptGenRandom(ghProvider, sizeof(WORD), (BYTE *)&result))
		{
			return result;
		}
	  }
    #else
	  if(fpRandom)
   	  {
		if(fread(&result, 1, sizeof(WORD), fpRandom))
		{
			return result;
		}
	  }
    #endif
	
	return (WORD) rand()%65536;
}

//************************************************************************************
// Returns a buffer of random bytes
//
//************************************************************************************

void getRandBytes(unsigned char *buffer, int size)
{
   #if defined(WIN32)
   	if(ghProvider)
	{
		if (CryptGenRandom(ghProvider, size, (BYTE *)buffer))
		{
			return;
		}
	}
   #else
	if(fpRandom)
	{
		if(fread(buffer, 1, size, fpRandom))
		{
			return;
		}
	}
   #endif
   for(int i = 0; i < size; i++)
	{
		buffer[i] = rand() % 256;
	}
}

