//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#if !defined(WIN32)
  #include "windowsdef.h"
#else
  #include <windows.h>
#endif

#include "base64.h"


unsigned char B64EncTable[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
char B64DecTable[256];
BOOL B64DecTableDone = FALSE;


//************************************************************************************
// Setup Base64 decoding table
//
//************************************************************************************

void B64SetupDecTable()
{
	if(!B64DecTableDone)
	{
		memset(B64DecTable, -1, sizeof(B64DecTable));

		short i;

		for(i = 'A'; i <= 'Z'; i++)
			B64DecTable[i] = i - 'A';

		for(i = 'a'; i <= 'z'; i++)
			B64DecTable[i] = 26 + i - 'a';

		for(i = '0'; i <= '9'; i++)
			B64DecTable[i] = 52 + i - '0';

		B64DecTable['+'] = 62;
		B64DecTable['/'] = 63;
	}
	B64DecTableDone = TRUE;
}


//************************************************************************************
// Returns B64 encoded size
//
//************************************************************************************

uint32_t B64GetEncSize(uint32_t inputSize)
{
	return (uint32_t)((double)(inputSize + 2) / 3) << 2;
}



//************************************************************************************
// Performs B64 encoding
//
// user must malloc the required size in output, otherwise...
//************************************************************************************

uint32_t B64Encode(unsigned char  *input, uint32_t inputSize,
                        unsigned char  *output)
{
	uint32_t outputSize = 0;

	if(output)
	{
		uint32_t src, dest;
		outputSize = B64GetEncSize(inputSize);

		for(src = 0, dest = 0; dest < outputSize; src += 3, dest += 4)
		{
			uint32_t val;
			BOOL secondChar = FALSE;
			BOOL thirdChar = FALSE;

			val = ((uint32_t)input[src] & 0xFF) << 8;

			if(src + 1 < inputSize)
			{
				val |= ((uint32_t)input[src + 1]) & 0xFF;
				secondChar = TRUE;
			}
			val <<= 8;

			if(src + 2 < inputSize)
			{
				val |= ((uint32_t)input[src + 2]) & 0xFF;
				thirdChar = TRUE;
			}

			output[dest + 3] = B64EncTable[(thirdChar ? (val & 0x3F) : 64)];
			val >>= 6;
			output[dest + 2] = B64EncTable[(secondChar ? (val & 0x3F) : 64)];
			val >>= 6;
			output[dest + 1] = B64EncTable[(val & 0x3F)];
			val >>= 6;
			output[dest] = B64EncTable[(val & 0x3F)];
		}
	}

	return outputSize;

}


//************************************************************************************
// Returns B64 decoded size
//
//************************************************************************************

uint32_t B64GetDecSize(uint32_t inputSize)
{
	return ((inputSize + 3) >> 2) * 3;
}

//************************************************************************************
// Performs B64 decoding
//
// user must malloc the required size in output, otherwise...
//************************************************************************************

uint32_t B64Decode(unsigned char  *input, uint32_t inputSize,
                        unsigned char  *output)
{
	uint32_t outputSize = 0;

	if(output)
	{
		unsigned short accum = 0;
		unsigned short bitsCount = 0;
		uint32_t outIndex = 0;

		B64SetupDecTable();

		for(uint32_t i = 0; i < inputSize; i++)
		{
			char inputChar = B64DecTable[input[i]];
			//printf("\n[%c] Decode = %.2X\n", input[i], inputChar);

			if(inputChar >= 0)
			{
				accum <<= 6;
				accum |= inputChar;
				bitsCount += 6;

				if(bitsCount >= 8)
				{
					bitsCount -= 8;
					output[outIndex++] = (unsigned char)((accum >> bitsCount) & 0xFF);
				}
			}
		}
		outputSize = outIndex;
	}

	return outputSize;
}


