//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#if !defined(WIN32)
  #include <string.h>
  #include <stdlib.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <arpa/inet.h>
  #include "windowsdef.h"
#else
  #include <windows.h>
#endif

#include "userRecord.h"

#include "rude/config.h"

using namespace rude;
char currentSection[30];
Config serverConfig;


const char *getConfigCurrentSection(void)
{
     return currentSection;
}
void setService(const char *service) 
{
     int len;
     char configSectionName[30];
     memset(configSectionName,0,30);
     strcpy(configSectionName,"Service:");
     len=strlen(service);
     if (len>20) len=20;
     strncpy(configSectionName+8,service,len);
     if (!serverConfig.setSection(configSectionName,FALSE))
     {     
           printf("There is no [%s] section in config file\n",configSectionName);
     }
     else
     {
         strncpy(currentSection,configSectionName,30);
     }
}

void setUserService(const char *user,unsigned short int serv)
{
     int len;
     char configSectionName[30],valueName[20];
     memset(configSectionName,0,30);
     strcpy(configSectionName,"User:");
     len=strlen(user);
     if (len>20) len=20;
     strncpy(configSectionName+5,user,len);
     sprintf(valueName,"Service:%d",serv);
     if (serverConfig.setSection(configSectionName,FALSE))
     {
         if (serverConfig.exists(valueName))
         {
            setService(serverConfig.getStringValue(valueName));
         }
         else
         {
            serverConfig.setSection(currentSection);
         }
     }
}

WORD getConfigUDPPort(void)
{
    WORD result;
    if (!serverConfig.exists("ListeningPort"))
    {
       serverConfig.setSection("Global");
    }   
    result=serverConfig.getIntValue("ListeningPort");
    serverConfig.setSection(currentSection);
    return result;
}

bool getConfigAllowMultipleConnections(void)
{
    WORD result;
    if (!serverConfig.exists("AllowMultipleConnections"))
    {
       serverConfig.setSection("Global");
    }   
    if (!serverConfig.exists("AllowMultipleConnections"))
    { result=FALSE;
    }
    else 
    {    result=serverConfig.getBoolValue("AllowMultipleConnections");
    }
    serverConfig.setSection(currentSection);
    return result;
}

bool getConfigClientCanSetPort(const char *user)
{
    bool result=FALSE;
    int len;
    char configSectionName[30];
    memset(configSectionName,0,30);
    strcpy(configSectionName,"User:");
    len=strlen(user);
    if (len>20) len=20;
    strncpy(configSectionName+5,user,len);
    if (serverConfig.setSection(configSectionName,FALSE))
    {
        if (serverConfig.exists("ClientCanSetPort"))
        {
           result=serverConfig.getBoolValue("ClientCanSetPort");
           serverConfig.setSection(currentSection);
        }
        else
        {
          serverConfig.setSection(currentSection);    
          if (serverConfig.exists("ClientCanSetPort"))
          {
            result=serverConfig.getBoolValue("ClientCanSetPort");
          }
        }
     }
    return result;
}
DWORD getConfigForwardToIP(void)
{
    DWORD result;
    if (!serverConfig.exists("ForwardToIP"))
    {
       serverConfig.setSection("Global");
    }   
    result=inet_addr(serverConfig.getStringValue("ForwardToIP"));
    serverConfig.setSection(currentSection);
    return result;
}

WORD getConfigForwardToPort(void)
{
    WORD result;
    if (!serverConfig.exists("ForwardToPort"))
    {
       serverConfig.setSection("Global");
    }   
    result=serverConfig.getIntValue("ForwardToPort");
    serverConfig.setSection(currentSection);
    return result;
}

WORD getConfigSingleConnPortOpenTime(void)
{
    WORD result;
    if (!serverConfig.exists("PortOpenTime"))
    {
       serverConfig.setSection("Global");
    }   
    result=serverConfig.getIntValue("PortOpenTime");
    serverConfig.setSection(currentSection);
    return result;
}

const char *getPktFilterInterface(void)
{
    const char *result;
    if (!serverConfig.exists("PktFilterInterface"))
    {
       serverConfig.setSection("Global");
    }   
    result=serverConfig.getStringValue("PktFilterInterface");
    serverConfig.setSection(currentSection);
    return result;
}

void printConfig(void)
{
		printf("Listening port = %d\n", getConfigUDPPort());
		printf("Forward to %s:%d\n", serverConfig.getStringValue("ForwardToIP"),getConfigForwardToPort());
		printf("Port open for %d sec.\n", getConfigSingleConnPortOpenTime());
}


bool loadServerConfig(char *filename)
{
    bool result;
    // Default configuration
    serverConfig.setSection("Global",TRUE);
    serverConfig.setIntValue("ListeningPort",1001);
    serverConfig.setStringValue("ForwardToIP","127.0.0.1");
    serverConfig.setIntValue("ForwardToPort",5900);
    serverConfig.setIntValue("PortOpenTime",30);
    serverConfig.setStringValue("PktFilterInterface","eth0");
    result=serverConfig.load (filename);
    serverConfig.setSection("Global",FALSE);
    if (serverConfig.exists("DefaultService"))
    {
       printf("Set default service : %s\n",serverConfig.getStringValue("DefaultService"));
       setService(serverConfig.getStringValue("DefaultService"));
   
    }
    return result;
}
