//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

// VERIFIER psGetUserRecord pour la partie Linux

#include <stdio.h>
#include "portForward.h"
#include "serverProcessPkt.h"
#include "md5.h"
#include "randomGen.h"
#include "aes.h"
#include "version.h"
#include "serverLog.h"
#include "serverConfig.h"
#include "aesCBC.h"
#include "platformSpecificFunc.h"
#if !defined(WIN32)
  #include <string.h>
  #include <malloc.h>
  #include "windowsdef.h"
  #include "privProcSupport.h"
  #include "unix/serverProcessPkt_unix.cpp"
#else
  #include "win32/serverProcessPkt_win32.cpp"
#endif
extern short int  listenAllPorts;  // Flag to listen to all port when receiving knocks

//************************************************************************************
// calculates ip, tcp, udp checksum
//
//************************************************************************************

unsigned int calculate_checksum(unsigned short* addr, int count)
{
    // Compute Internet Checksum for "count" bytes
    // beginning at location "addr".
    
	uint32_t sum = 0;

    while( count > 1 )
	{
		// This is the inner loop
		sum += * ((unsigned short*) addr++);
		count -= 2;
	}

	// Add left-over byte, if any
	if( count > 0 )
		sum += * (unsigned char *) addr;

    //  Fold 32-bit sum to 16 bits
	while (sum>>16)
		sum = (sum & 0xffff) + (sum >> 16);

	return (unsigned int) ~sum;
}


//************************************************************************************
// Returns "size" padded up to 16 byte boundary.
// If size is already 0 mod 16, then return size + 16.
//
//************************************************************************************

int getPaddedSize(int size)
{
	int result = (((size - 1) >> 4) + 1) << 4;		// aligned to 16 bytes
	if(result == size)
		result += 16;

	return result;
}


//************************************************************************************
// Decrypts P1 and P2 packet from server
//
//************************************************************************************

BOOL decryptUDPReq(char *requestData, int size, char **userid, LPVOID *decPktData, 
				   unsigned short int *pktType)
{
	if((size - SIZEOF_USERID) & 0xF)		// size % 16
    {
    }
	char *id = requestData;
		
	if(id[SIZEOF_USERID-1] != 0)		// userid has max 8 char and must be NULL terminated
	{
//		printf("Rejected in decryptUDPReq 1\n");
		return FALSE;
	}
	
	USER_RECORD userRec;

	
	#if !defined(WIN32)
	  if(!psGetUserRecord(id, &userRec))
	#else
	  if(!getUserRecord(USER_FILENAME, id, &userRec))
	#endif
	{
//		printf("User record not found!\n");
		return FALSE;
	}

	// AES decrypt code
	PORTSEQ_REQ *pq = (PORTSEQ_REQ *)(requestData + SIZEOF_USERID);	
	unsigned char encKey[16];
	genEncKey(userRec.userName, userRec.userPasswordHash, encKey);

	aes_ctx aesState;
	aes_dec_key(encKey, sizeof(encKey), &aesState);
	aesDecrypt(&aesState, (unsigned char *)pq, size - SIZEOF_USERID);

	// version check
	if(pq->version != VERSION)	// version is single byte
	{
//		printf("****Version Rejected (%d)\n", pq->version);
		return FALSE;
	}

	WORD recvPktType = ntohs(pq->packetType);
	// packet type check
	if(recvPktType != P1 && recvPktType != P2 && recvPktType != PK)
	{
//		printf("Rejected in decryptUDPReq 4 (%x)\n", recvPktType);
		return FALSE;
	}

	// MD5 check
	md5byte recvHash[16], digest[16];
    	int i;
	for(i = 0; i < 16; i++)
	{
		recvHash[i] = pq->MD5[i];
		pq->MD5[i] = 0;
	}

	MD5Context ctx;
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *)userRec.userName, strlen(userRec.userName));
	MD5Update(&ctx, (unsigned char *)userRec.userPasswordHash, SIZEOF_USERPASSWORD_HASH);
	if(recvPktType == P1)
		MD5Update(&ctx, (unsigned char *)pq, sizeof(PORTSEQ_REQ));
	else if(recvPktType == P2)
		MD5Update(&ctx, (unsigned char *)pq, sizeof(PORTREQ_QUERY));
	else if(recvPktType == PK)
		MD5Update(&ctx, (unsigned char *)pq, sizeof(PORTKNOCK));



	MD5Final(digest, &ctx);
	if(memcmp(digest, recvHash, 16) != 0)
	{
//		printf("Hash Error\n");
		return FALSE;
	}

	
	// BOTH P1 and P2 have version and packetType at same pos
	pq->packetType = ntohs(pq->packetType);		
	// if type == P1 check lastTime
	if(pq->packetType == P1)
	{
		// convert from network to host byte order	
		pq->reqType = ntohs(pq->reqType);
		pq->timeStamp = ntohl(pq->timeStamp);
        pq->service=ntohs(pq->service);
        pq->portLow=ntohs(pq->portLow);
        pq->portHigh=ntohs(pq->portHigh);
		if(pq->timeStamp < userRec.lastTime)
		{
//			printf("Replay Rejected\n");
			return FALSE;
		}
		
		for(i = 0; i < MAX_KNOCKS; i++)
			pq->ports[i] = ntohs(pq->ports[i]);

		for(i = 0; i < MAX_KNOCKS; i++)
			pq->ISNs[i] = ntohl(pq->ISNs[i]);
	}
	else if(pq->packetType == P2)
	{
		PORTREQ_QUERY *prq = (PORTREQ_QUERY *)pq;

		// convert from network to host byte order	
		prq->reqType = ntohs(prq->reqType);
		prq->timeStamp = ntohl(prq->timeStamp);
		if(prq->timeStamp < userRec.lastTime)
		{
//			printf("Replay Rejected\n");
			return FALSE;
		}

	}
	else if(pq->packetType == PK)
	{
		PORTKNOCK *pk= (PORTKNOCK *) pq;
		pk->reqType = ntohs(pk->reqType);
		pk->timeStamp = ntohl(pk->timeStamp);
		pk->ISN=ntohl(pk->ISN);
		if(pk->timeStamp < userRec.lastTime)
		{
//			printf("Replay Rejected %ul %ul\n",pk->timeStamp,userRec.lastTime);
			return FALSE;
		}
		//printf("Packet PK ISN %.8X\n",pk->ISN);
	}

	*decPktData = (LPVOID)pq;
	*userid = id;
	*pktType = pq->packetType;
	return TRUE;
}



//************************************************************************************
// Returns pointer to the UDP header.  The purpose of this function is to reject as
// many invalid packets as possible to reduce the amount of processing later on.
//
// Only packets that passes all the conditions will be subsequently decrypted.  This
// tries to prevent computation attacks on server.
//
//************************************************************************************

LPVOID getUDPData(char *pkt, int size, struct in_addr *serverIPAddr,
				  IP_HEADER **out_iphdr, UDP_HEADER **out_udphdr, int *outDataSize)
{
	int minHeaderSize = sizeof(IP_HEADER) + sizeof(UDP_HEADER);
	if(size <= minHeaderSize)
		return NULL;

	IP_HEADER *iphdr;

	iphdr = (IP_HEADER *)pkt;
	int iphdrLen = iphdr->ip_hl << 2;
	if(iphdrLen >= size)
		return NULL;
	
	//printf("IP Protocol Check (%X)\n", iphdr->ip_p);
	if(iphdr->ip_p != IPPROTO_UDP)
		return NULL;

	//printf("->IP Check\n");
  	if(iphdr->ip_dst != dk_ipaddr(serverIPAddr))
		return NULL;

	//printf("->IP Len Check\n");
	if(size - iphdrLen <= sizeof(UDP_HEADER))
		return NULL;

	UDP_HEADER *udphdr;
	udphdr = (UDP_HEADER *)(pkt + iphdrLen);
	
	if((!listenAllPorts)&&(ntohs(udphdr->th_dport) != getConfigUDPPort()))
	{
//        printf("Wrong port : %d\n",getConfigUDPPort());
		return NULL;
     }
	int dataSize = size - iphdrLen - sizeof(UDP_HEADER);
	if(size <= 0)
		return NULL;
	
	if((dataSize != getPaddedSize(sizeof(PORTSEQ_REQ)) + SIZEOF_USERID) &&
	   (dataSize != getPaddedSize(sizeof(PORTREQ_QUERY)) + SIZEOF_USERID) &&
       (dataSize != getPaddedSize(sizeof(PORTKNOCK)) + SIZEOF_USERID))		// data must be multiple of 16 bytes
       {
        return NULL;
        }
	if ((dataSize > 150)||(dataSize<16)) 
	{
// 		printf("->Data Size Boundary Reject\n");
		return NULL;
	}



	*out_iphdr = iphdr;
	*out_udphdr = udphdr;
	*outDataSize = dataSize;

	return (pkt + iphdrLen + sizeof(UDP_HEADER));
}


//************************************************************************************
// Processes TCP SYN knock from client.
//
//************************************************************************************

BOOL getTCPKnockData(char *pkt, int size, struct in_addr *serverIPAddr,
					 DWORD *ISN, WORD *port, struct in_addr *srcIP)
{
	int minHeaderSize = sizeof(IP_HEADER) + sizeof(TCP_HEADER);
	if(size < minHeaderSize)
	{
		return FALSE;
	}

	IP_HEADER *iphdr;

	iphdr = (IP_HEADER *)pkt;
	int iphdrLen = iphdr->ip_hl << 2;
	if(iphdrLen >= size)
	{
		return FALSE;
	}
	
	if(iphdr->ip_p != IPPROTO_TCP)
	{
		return FALSE;
	}

	// ipaddr check
    if(iphdr->ip_dst != dk_ipaddr(serverIPAddr))
	{
		return FALSE;
	}

	if(size - iphdrLen < sizeof(TCP_HEADER))
	{
		return FALSE;
	}

	TCP_HEADER *tcphdr;
	tcphdr = (TCP_HEADER *)(pkt + iphdrLen);

	if(tcphdr->th_flags != TH_SYN)
	{
		return FALSE;
	}

	*port = ntohs(tcphdr->th_dport);
	*ISN = ntohl(tcphdr->th_seq);
	dk_setaddr(srcIP,iphdr->ip_src);
	
	return TRUE;
}


//************************************************************************************
// Returns the protocol type of the IP header.
//
//************************************************************************************

DWORD getIPPktType(char *pkt, int size)
{
	int minHeaderSize = sizeof(IP_HEADER) + sizeof(UDP_HEADER);
	if(size <= minHeaderSize)
		return 0; // return 0?

	IP_HEADER *iphdr;

	iphdr = (IP_HEADER *)pkt;
	return iphdr->ip_p;
}


//************************************************************************************
// Returns source IP address of an IP packet.
//
//************************************************************************************

DWORD getIPSrcAddr(char *pkt, int size)
{
	int minHeaderSize = sizeof(IP_HEADER) + sizeof(UDP_HEADER);
	if(size <= minHeaderSize)
		return 0; //return 0

	IP_HEADER *iphdr;

	iphdr = (IP_HEADER *)pkt;
	return iphdr->ip_src;
}


//************************************************************************************
// Extracts the IP packet from an Ethernet II packet
//
//************************************************************************************

LPVOID getIPHeader(char *packet, int size, int *outIPPktLen)
{
	if(size <= ETH2_HEADER_SIZE)
		return NULL;

	// we support only ETH2 packets in this version
	ETH2_HEADER *e2Hdr = (ETH2_HEADER *)packet;
	if(e2Hdr->pktType == ETHPROTO_IP)
	{
		*outIPPktLen = size - sizeof(ETH2_HEADER);
		return (LPVOID)(packet + ETH2_HEADER_SIZE); 
	}

	return NULL;
}


//************************************************************************************
// Returns the source port of an UDP packet.
//
//************************************************************************************

WORD getUDPSrcPort(char *pkt, int size)
{
	int minHeaderSize = sizeof(IP_HEADER) + sizeof(UDP_HEADER);
	if(size <= minHeaderSize)
		return 0; // return 0

	IP_HEADER *iphdr;

	iphdr = (IP_HEADER *)pkt;

	UDP_HEADER *udphdr;
	int iphdrLen = iphdr->ip_hl << 2;

	udphdr = (UDP_HEADER *)(pkt + iphdrLen);

	return ntohs(udphdr->th_sport);
}


//************************************************************************************
// Creates a P3 packet to reply to client.
//
//	unsigned char version;	
//	unsigned short int packetType;		// P3
//	unsigned short int reqType;			// single connection/multiple connections
//	unsigned short int allocatedPort;
//	unsigned short int portOpenTime;	// port open time
//	uint32_t timeStamp;
//
//************************************************************************************

char *makePortSeqQueryReplyPkt(WORD allocPort, uint32_t timeStamp, char *userID, 
							   unsigned char *userPasswordHash, WORD reqType, int *outPktSize)
{
	PORTREQ_QUERY_REPLY prqr;
	memset(&prqr, 0, sizeof(PORTREQ_QUERY_REPLY));
	
	// filled up fields in network order
	prqr.allocatedPort = htons(allocPort);
	prqr.packetType = htons(P3);
	prqr.version = VERSION;
	prqr.timeStamp = htonl(timeStamp);
	prqr.portOpenTime = htons(getConfigSingleConnPortOpenTime());
	prqr.reqType = htons(reqType);

	getRandBytes(prqr.IV, 16);

	// Calc MD5
	MD5Context ctx;
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *)userID, strlen(userID));
	MD5Update(&ctx, (unsigned char *)userPasswordHash, SIZEOF_USERPASSWORD_HASH);
	MD5Update(&ctx, (unsigned char *)&prqr, sizeof(PORTREQ_QUERY_REPLY));

	md5byte digest[16];
	MD5Final(digest, &ctx);

	for(int i = 0; i < 16; i++)
		prqr.MD5[i] = digest[i];

	int pktSize;
	char *result = padAndEncrypt(userID, userPasswordHash, (char *)&prqr, 
		                         sizeof(PORTREQ_QUERY_REPLY), &pktSize);

	if(result)
		*outPktSize = pktSize;

	return result;
}




//************************************************************************************
// Returns a buffer containing data padded up to boundary of 16 bytes and AES
// encrypted using userID and password hash.
//
// This buffer must be freed after use.
//
//************************************************************************************

char *padAndEncrypt(char *userID, unsigned char *userPasswordHash, char *data, int size, int *outSize)
{
	int paddedSize = getPaddedSize(size);
	int reqSize;

	if(paddedSize == size)
		reqSize = size + 16;
	else 
		reqSize = paddedSize;

	char *result = (char *)malloc(reqSize);
	if(result)
	{
		memcpy(result, data, size);
		int padLen = reqSize - size;

		unsigned char *ptr = (unsigned char *)result + size;
		for(int i = size; i < reqSize; i++)
		{
			result[i] = (unsigned char)padLen;
		}

		unsigned char encKey[16];
		genEncKey(userID, userPasswordHash, encKey);

		aes_ctx aesState;
		aes_enc_key(encKey, sizeof(encKey), &aesState);
		aesEncrypt(&aesState, (unsigned char *)result, reqSize);

		*outSize = reqSize;

		memset(&aesState, 0, sizeof(aes_ctx));
		memset(encKey, 0, sizeof(encKey));		// zero out
	}
	return result;
}
