//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************




//************************************************************************************
// Prints a list of pcap interfaces and usage instructions
//
//
//************************************************************************************

void printUsage(char *exeName)
{
	struct ifconf   ifc;
	struct ifreq    ibuf[32], ifr, *ifrp, *ifend;
	SOCKET fd;
					
	if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) 
		goto done;
		
	memset(ibuf, 0, sizeof(struct ifreq) << 5);
	ifc.ifc_len = sizeof ibuf;
	ifc.ifc_buf = (caddr_t) ibuf;

	// gets interfaces list
	if (ioctl(fd, SIOCGIFCONF, (char*)&ifc) == -1 ||
		ifc.ifc_len < sizeof(struct ifreq))
	{
		close(fd);
		goto done;
	}
	
	// ifrp points to buffer and ifend points to buffer's end
	ifrp = ibuf;
	ifend = (struct ifreq*) ((char*)ibuf + ifc.ifc_len);

	for (; ifrp < ifend; ifrp++) 
	{
		memset(ifr.ifr_name, 0, sizeof(ifr.ifr_name));		
		strncpy(ifr.ifr_name, ifrp->ifr_name, sizeof(ifr.ifr_name)-1);
		
		if (ioctl(fd, SIOCGIFFLAGS, (char*)&ifr) == -1) 
			continue;

		// Down interface?
		if (!(ifr.ifr_flags & IFF_UP))
		{
			printf("%-8s - interface down\n", ifr.ifr_name);
			continue;
		}
		// Get the interface address
		if (ioctl(fd, SIOCGIFADDR, (char*)&ifr) == -1) 
			continue;

		struct sockaddr_in *sin;
			
		sin = (sockaddr_in *) &(ifr.ifr_addr);
		printf("%-8s - %s\n", ifr.ifr_name, inet_ntoa(sin->sin_addr));
	}
done:
	printf("\nUsage: %s -i <interface>", exeName);
	printf("\n       %s -i <interface> -D    (to run as daemon)", exeName);
        printf("\n       %s -i <interface> -D -S (to run as daemon and use syslog)\n", exeName);
	printf("\nIf syslog is not used, program logs to %s when running as daemon.\n\n", SERVER_LOGFILE);
}













//************************************************************************************
// Returns the IP address of the interface
//
//************************************************************************************

DWORD getInterfaceIP(char *ifname)
{
	int fd = socket(AF_INET, SOCK_DGRAM, 0);
	if(fd != INVALID_SOCKET)
	{
		struct ifreq ifr;
		strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name)); 
		
		if(ioctl(fd, SIOCGIFADDR, (char *)&ifr) < 0)
		{
			return 0;
		}
		else
		{
			struct sockaddr_in *sin;
			
			sin = (sockaddr_in *) &(ifr.ifr_addr);
			return sin->sin_addr.s_addr;
		}
	}

	return 0;
}


//************************************************************************************
// Handle program termination
// 
//************************************************************************************

void sigParentExit(int signum)
{
	if(!gTerminateApp)
	{
		gTerminateApp = TRUE;

		logString("Server Terminated.\n");
		if(gDaemon)
			unlink(PID_FILE);
	
		kill(gPktProcPid, SIGTERM);
		endFwSupport();
		endLog();
	}
	else
		return;
	exit(signum);
}


//************************************************************************************
// Handle child termination
// 
//************************************************************************************

void sigChild(int signum)
{
	wait(NULL);
	return;
}


//************************************************************************************
// Program entry
//
// This version uses normal UDP sockets to send outgoing UDP port sequence requests
// so firewall rules must allow outgoing UDP packets.  i.e.
//
// iptables -A OUTPUT -p udp -j ACCEPT
//
//************************************************************************************

int main(int argc, char* argv[])
{
	printf("DrKnockd Version %s Copyright (c) 2006 Olivier ROCH VILATO and Chew Keong TAN\n\n", PROG_VERSION);

	int opt;
	char *interface = NULL;

	if(argc < 3)
	{
		printUsage(argv[0]);
		return 1;
	}

	while((opt = getopt(argc, argv, "i:DS")) != -1)
	{
		switch(opt)
		{
			case 'i':
				interface = optarg;
				break;

			case 'D':
				gDaemon = TRUE;
				break;
			case 'S':
				gUseSyslog = TRUE;
				break;
			default:
				break;
		}
	}

	if(!interface)
	{
		printUsage(argv[0]);
		return 1;
	}

	struct passwd *pw;
	pw = getpwnam("drknock");
	if(!pw)
	{
		printf("Non-privileged user \"drknock\" is required but not found on this system!\n"
		       "Add this user to /etc/passwd. This user's home directory will be used\nfor chroot.\n\n");
		return 1;
	}
	DIR *testDir;
	testDir = opendir(pw->pw_dir);
	if(!testDir)
	{
		printf("Home directory for non-privileged user \"drknock\" is invalid!\n\n");
		return 1;
	}
	closedir(testDir);

	gDropId = pw->pw_uid;
	gDropGrp = pw->pw_gid;

	FILE *fileChk = fopen(USER_FILENAME, "r");
	if(fileChk == NULL)
	{
		printf("You must first create a user account file (%s), and use\n"
		       "drknockd_useradd to create accounts.\n\n",
			USER_FILENAME);
		return 1;
	}
	fclose(fileChk);

	if(!initRand())
	{
		printf("Error initializing random generator!\n");
		return 1;
	}

	if(!loadServerConfig(CONFIG_FILENAME))
	{
		printf("Configuration file <%s> not found! Using defaults.\n", CONFIG_FILENAME);
	}

	char errbuf[PCAP_ERRBUF_SIZE];
   
	gOurIPAddr.s_addr = getInterfaceIP(interface);
	if(gOurIPAddr.s_addr == 0)
	{
		printf("Interface is invalid!\n\n");
		return 1;
	}
	strcpy(gOurIPAddrStr, inet_ntoa(gOurIPAddr));
   
	printConfig();
	printf("\nOpening %s (%s)\n", interface, gOurIPAddrStr);

	if ( (adhandle = pcap_open_live(interface,		// name of the device
				1000,				// portion of the packet to capture. 
				TRUE,				// promiscuous mode
				500,				// read timeout
				errbuf				// error buffer
				) ) == NULL)
	
	{
		fprintf(stderr,"\nUnable to open the adapter. Make sure you are running this as root.\n");        
		return 1;
	}

	if(gDaemon)
	{
		printf("Server started as daemon.\n");
		if(gUseSyslog)
			printf("Logging to syslog.\n\n");
		else
			printf("\n");
		
		if(fork() > 0)		// parent process
		{
			pcap_close(adhandle);
			adhandle = NULL;
			_exit(0);
		}
		else
		{
			setsid();
			close(0);			
			close(1);
			close(2);
			initLog(SERVER_LOGFILE);
	        	
			// write PID to file
			FILE *fp;
			if((fp = fopen(PID_FILE, "w")))
			{
				fprintf(fp, "%d\n", getpid());
				fclose(fp);
			}
		}
	}
	else
	{
		printf("Server started.\n");
		if(gUseSyslog)
			printf("Logging to syslog.\n\n");
		else
			printf("\n");
		initLog(NULL);		// log to stdout
	}
	
	// write PID to file
	FILE *fp;
	if((fp = fopen(PID_FILE, "w"))) 
	{
		fprintf(fp, "%d\n", getpid());
		fclose(fp);
	}
			
	initPrivProcSupport();
	
	gPktProcPid = fork();
	if(gPktProcPid == 0)	// child
	{
		chdir(pw->pw_dir);
		chroot(pw->pw_dir);
		setregid(gDropGrp, gDropGrp);
		setreuid(gDropId, gDropId);		// drop priv

		signal(SIGCHLD, sigChild);
		psLowerPrivClosePipes();		// close unnecesary end of priv support pipe

		initList(&gPortSeqRequestList);
		sleep(1);
		
		pcap_loop(adhandle, 0, packet_handler, NULL);
		_exit(0);
	}

	// parent signal handlers
	signal(SIGCHLD, sigChild);
    signal(SIGINT, sigParentExit);
    signal(SIGTERM, sigParentExit);

    logString("Server started.\n");

	pcap_close(adhandle);	
	
	startPrivProcSupport();
	privProcThreadFunc(NULL);

	return 0;
}

