//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <semaphore.h>
#include <pthread.h>
#include <unistd.h>

#include <portSeqReqList.h>
#include <randomGen.h>
#include <windowsdef.h>
#include <serverLog.h>
#include <portForward.h>
#include <serverConfig.h>
#include <firewallSupport.h>


int		gFwSupportPipe[2];
LINKED_LIST	gFwSupportList;
extern BOOL	gTerminateApp;
sem_t		gFwSupportSem;


//************************************************************************************
// Executes iptables to close firewall after client has disconnected
//
//************************************************************************************

BOOL auxCloseFirewall(DWORD trustedIP, WORD allocPort)
{
	char *args[] = { "iptables", 
			"-D", "INPUT", 
			"-s", "192.168.1.3",
			"-p", "tcp",
			"-m", "tcp",
			"--dport", "22",
			"-m", "state",
			"--state", "NEW,ESTABLISHED",
			"-j", "ACCEPT",
			NULL };
	
	char portStr[20];
	struct in_addr ina;
	ina.s_addr = trustedIP;

	snprintf(portStr, sizeof(portStr) - 1, "%u", allocPort);
	args[4] = inet_ntoa(ina);
	args[10] = portStr;

	pid_t pid;
	if((pid = vfork()) == 0)
	{
		execvp("iptables", args);
		_exit(0);
	}
	else
		waitpid(-1, NULL, WNOHANG);

	sleep(1);
	args[2] = "OUTPUT";
	args[3] = "-d";
	args[9] = "--sport";
	args[14] = "ESTABLISHED";
	
	if((pid = vfork()) == 0)
	{
		execvp("iptables", args);
		_exit(0);
	}
	else
		waitpid(-1, NULL, WNOHANG);
	
	return TRUE;
}


//************************************************************************************
// Executes iptables to open firewall for accepted clients
//
// iptables -A INPUT -s 192.168.1.3 -p tcp -m tcp --dport 22 -m state 
//          --state NEW,ESTABLISHED -j ACCEPT
//
//************************************************************************************

BOOL auxOpenFirewall(DWORD trustedIP, WORD allocPort)
{
	char *args[] = { "iptables", 
			"-A", "INPUT", 
			"-s", "192.168.1.3",
			"-p", "tcp",
			"-m", "tcp",
			"--dport", "22",
			"-m", "state",
			"--state", "NEW,ESTABLISHED",
			"-j", "ACCEPT",
			NULL };	
	char portStr[20];
	struct in_addr ina;
	ina.s_addr = trustedIP;

	snprintf(portStr, sizeof(portStr) - 1, "%u", allocPort);
	args[4] = inet_ntoa(ina);
	args[10] = portStr;

	pid_t pid;
	if((pid = vfork()) == 0)
	{
		execvp("iptables", args);
		_exit(0);
	}
	else
		waitpid(-1, NULL, WNOHANG);
	
	args[2] = "OUTPUT";
	args[3] = "-d";
	args[9] = "--sport";
	args[14] = "ESTABLISHED";

	sleep(1);	
	if((pid = vfork()) == 0)
	{
		execvp("iptables", args);
		_exit(0);
	}
	else
		waitpid(-1, NULL, WNOHANG);
	
	return TRUE;
}


//************************************************************************************
// Called by the privilege part of this program to add a firewall rule.
// This function will also add the record to a list, so that later on, the rule
// can be removed.
//
//************************************************************************************

BOOL fwOpenFirewall(DWORD trustedIP, WORD allocPort)
{
	FIREWALL_SUPPORT *newFw = (FIREWALL_SUPPORT *)malloc(sizeof(FIREWALL_SUPPORT));
	BOOL result = FALSE;
	
	if(newFw)
	{
		newFw->allowedIP = trustedIP;
		newFw->portNum = allocPort;
		
		sem_wait(&gFwSupportSem);
		BOOL inst = insertItem(&gFwSupportList, (LPVOID)newFw);

		if(!inst)
			free(newFw);
		else
		{
			auxOpenFirewall(trustedIP, allocPort);
			result = TRUE;
		}
		sem_post(&gFwSupportSem);
	}
	return result;
}


//************************************************************************************
// Called by the privilege part of this program to remove a firewall rule.
// This function will also remove the record from the list, 
//
//************************************************************************************

BOOL fwCloseFirewall(DWORD trustedIP, WORD allocPort)
{
	FIREWALL_SUPPORT fw;
	BOOL result = FALSE;

	fw.allowedIP = trustedIP;
	fw.portNum = allocPort;

	sem_wait(&gFwSupportSem);
	
	LPVOID data = deleteSingleItem(&gFwSupportList, fwCloseListFunc, (LPVOID)&fw);

	if(data)
	{
		free(data);
		auxCloseFirewall(fw.allowedIP, fw.portNum);
		result = TRUE;
	}
	sem_post(&gFwSupportSem);

	return result;
}


//************************************************************************************
// This is a callback function that attempts to remove all firewall rules that we
// have added.  But seems like it doesn't work now.
//
//************************************************************************************

BOOL fwCloseALLFunc(LPVOID data, LPVOID toFind)
{
	FIREWALL_SUPPORT *fw = (FIREWALL_SUPPORT *)data;

	//printf("close %d\n", fw->portNum);
	auxCloseFirewall(fw->allowedIP, fw->portNum);
	free(data);

	return TRUE;
}


//************************************************************************************
// This function is call by the main thread to remove all firewall rules that we
// have added.  But seems like it doesn't work now.
//
//************************************************************************************

void fwCloseALL(void)
{
	sem_wait(&gFwSupportSem);

	deleteMultipleItems(&gFwSupportList, fwCloseALLFunc, NULL);
	
	sem_post(&gFwSupportSem);
}


//************************************************************************************
// Callback function to find the firewall entry that contains details of the rule to
// remove from iptables.
// 
//************************************************************************************

BOOL fwCloseListFunc(LPVOID data, LPVOID toFind)
{
	FIREWALL_SUPPORT *fw = (FIREWALL_SUPPORT *)data;
	FIREWALL_SUPPORT *toFindFw = (FIREWALL_SUPPORT *)toFind;

	if(fw->allowedIP == toFindFw->allowedIP &&
	   fw->portNum == toFindFw->portNum)
	{
		return TRUE;
	}
	return FALSE;
}



void startFwSupport(void)
{
	initList(&gFwSupportList);
	sem_init(&gFwSupportSem, 0, 1);

	return;
}


void endFwSupport(void)
{
	fwCloseALL();
	sem_destroy(&gFwSupportSem);
	return;
}
