//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "randomGen.h"
#include "windowsdef.h"
#include "serverLog.h"
#include "portForward.h"
#include "serverConfig.h"
#include "privProcSupport.h"

extern BOOL		gTerminateApp;
extern uid_t		gDropId;
extern gid_t		gDropGrp;



//************************************************************************************
// Function to perform port forwarding. 
//
//************************************************************************************

DWORD forwardFunc(SOCKET client, struct in_addr ina, WORD portNum)
{	
	char buffer[65536];

	SOCKET f_sock = socket(AF_INET, SOCK_STREAM, 0);
	if(f_sock == INVALID_SOCKET)
		goto done;

	struct sockaddr_in sin;
	sin.sin_family = AF_INET;
	sin.sin_port = htons(getConfigForwardToPort());
	sin.sin_addr.s_addr = getConfigForwardToIP();

	if(connect(f_sock, (struct sockaddr *)&sin, sizeof(sin)) == SOCKET_ERROR)
	{
		logString("Connection to forwarded IP failed!\n");
		close(f_sock);		
	}
	else
	{
		while(true && !gTerminateApp)
		{
		        struct pollfd pfd[2];
	        	pfd[0].fd = client;
        		pfd[0].events = POLLIN | POLLPRI;
	        	pfd[1].fd = f_sock;
        		pfd[1].events = POLLIN | POLLPRI;

			int iEvent = poll(pfd, 2, -1);		// -1 infinite poll
			
			if(iEvent <= 0)
				break;

			// read event from "client" socket 
			if((pfd[0].revents & POLLIN) || (pfd[0].revents & POLLPRI))
			{
				int lBytesRead = recv(client, buffer, sizeof(buffer), 0);

				if(lBytesRead > 0)
				{
					send(f_sock, buffer, lBytesRead, 0);
				}
				else
					break;
			}
			if((pfd[0].revents & POLLERR) || (pfd[0].revents & POLLHUP))
			{
				break;	
			}
			
			// read event from f_sock socket
			if((pfd[1].revents & POLLIN) || (pfd[1].revents & POLLPRI))
			{
				int lBytesRead = recv(f_sock, buffer, sizeof(buffer), 0);

				if(lBytesRead > 0)
				{
					send(client, buffer, lBytesRead, 0);
				}
				else
					break;
			}
			if((pfd[1].revents & POLLERR) || (pfd[1].revents & POLLHUP))			
			{
				break;	
			}
		}
		close(f_sock);
	}

done:
	logString("Client %s disconnected from port %d.\n", inet_ntoa(ina), portNum);
	close(client);
	
	return 0;
}


//************************************************************************************
// Thread function to serve accepted clients.  In this version, we support only
// forwarding of one connection per request. i.e. The allocated port will only close
// after the first connection from the client is served.
//
//************************************************************************************

LPVOID forwarderThreadFunc(LPVOID data)
{
	//setreuid(gDropId, gDropId);	// drop priv
	//setregid(gDropGrp, gDropGrp);
	
	FORWARDER_DATA *fdata = (FORWARDER_DATA *)data;
	SOCKET s = fdata->s;
	DWORD allowedIP = fdata->allowedIP;
	WORD portNum = fdata->portNum;
	pthread_t pt = fdata->pt;

	free(data);
	struct sockaddr_in sin;

	DWORD portOpenTime = getConfigSingleConnPortOpenTime();		// in secs
	bool allowMultipleConnections=getConfigAllowMultipleConnections();
	DWORD startTime = time(NULL);
	DWORD timeLeft;

	psOpenFirewall(allowedIP, portNum);		// open firewall
	
	while(true && !gTerminateApp)
	{
	        struct pollfd pfd;
        	pfd.fd = s;
        	pfd.events = POLLIN;
        	
		timeLeft = portOpenTime - (time(NULL) - startTime);
	        
		if(poll(&pfd, 1, timeLeft * 1000) <= 0)
		{
			struct in_addr ina1;
			ina1.s_addr = allowedIP;

			if(!gTerminateApp)
				logString("Port %d for %s timeout and closed.\n", portNum, inet_ntoa(ina1));
			goto done;
		}		

		if(pfd.revents & POLLIN)
		{
			socklen_t sin_size = sizeof(sin);
			SOCKET client = accept(s, (struct sockaddr *)&sin, &sin_size);

			if(client != INVALID_SOCKET)
			{
				if(sin.sin_addr.s_addr != allowedIP)
				{
					// unauthorized connection
					logString("Unauthorized connection to port %d from %s!\n", 
						       portNum, inet_ntoa(sin.sin_addr));

					close(client);
					continue;
				}
				else
				{
					logString("Client from %s connected to port %d.\n", 
						      inet_ntoa(sin.sin_addr), portNum);
					forwardFunc(client, sin.sin_addr, portNum);
				}
				if (!allowMultipleConnections) 
				{ goto done; } 
				
			}
			else
			{
				logString("Socket is invalid when accepting connection from client.\n");
			}
		}
	}

done:
	psCloseFirewall(allowedIP, portNum);		// open firewall
	close(s);
	
	return NULL;
}
