//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <semaphore.h>
#include <pthread.h>
#include <unistd.h>
#include <syslog.h>

#include <windowsdef.h>
#include <firewallSupport.h>
#include <privProcSupport.h>
#include <userRecord.h>
#include <serverLog.h>

int		gPrivProcCommandPipe[2];
int		gPrivProcReplyPipe[2];
extern BOOL	gTerminateApp;

//************************************************************************************
// Creates the necessary pipes to communicate with the lower priv child process
//
//************************************************************************************

void initPrivProcSupport(void)
{
	pipe(gPrivProcCommandPipe);
	pipe(gPrivProcReplyPipe);
}


//************************************************************************************
// Initializes the firewall support
//
//************************************************************************************

void startPrivProcSupport(void)
{
	startFwSupport();
}


//************************************************************************************
// Called from the lower privilege child to close unnecessary pipes
//
//************************************************************************************

void psLowerPrivClosePipes(void)
{
	close(gPrivProcCommandPipe[0]);
	close(gPrivProcReplyPipe[1]);
}


//************************************************************************************
// Handles receiving of data from pipe.
//
//************************************************************************************

LPVOID psGetPipeReply(DWORD expSize)
{
	ssize_t recv = 0;
	char *reply = (char *)malloc(expSize);
	
	if(reply)
	{
	       	while(recv < expSize)
		{
			ssize_t curRecv = read(gPrivProcReplyPipe[0],
					       	reply + recv,
						expSize - recv);
			if(curRecv <= 0)
				goto done;
			else
				recv += curRecv;
		}
	}
done:
	if(reply && recv != expSize)
	{
		free(reply);
		reply = NULL;
	}
	return reply;
}


//************************************************************************************
// This function is meant to be used by lower priv child to retrieve user records
// from the account file.
//
//************************************************************************************
 
BOOL psGetUserRecord(char *userName, USER_RECORD *userRec)
{
	PRIVPROC_SUPPORT request;
	PRIVPROC_SUPPORT *reply;
	BOOL result = FALSE;

	memset(&request, 0, sizeof(request));
	request.type = GETUSER;
	strncpy(request.data.userRecord.userName, userName, SIZEOF_USERID - 1);

	write(gPrivProcCommandPipe[1], &request, sizeof(request));

	reply = (PRIVPROC_SUPPORT *)psGetPipeReply(sizeof(PRIVPROC_SUPPORT));
	if(reply)
	{
		// this will not work if there are multiple processes sending request at same time.
		if(reply->type == GETUSER_REPLY)
		{
			//printf("<%s>\n", reply->data.userRecord.userName);
			if(strlen(reply->data.userRecord.userName) > 0)
			{
				*userRec = reply->data.userRecord;
				result = TRUE;
			}
		}
		free(reply);
	}
	
	return result;
}



//************************************************************************************
// This function is meant to be called by the lower priv child to update the user's
// timestamp.
//
//************************************************************************************

BOOL psUpdateUserTimestamp(char *userName, DWORD newTimestamp)
{
	PRIVPROC_SUPPORT request;
	PRIVPROC_SUPPORT *reply;
	BOOL result = FALSE;
	
	memset(&request, 0, sizeof(request));
	request.type = SAVEUSER;
	strncpy(request.data.userRecord.userName, userName, SIZEOF_USERID - 1);
	request.data.userRecord.lastTime = newTimestamp;

	write(gPrivProcCommandPipe[1], &request, sizeof(request));
	
	reply = (PRIVPROC_SUPPORT *)psGetPipeReply(sizeof(PRIVPROC_SUPPORT));
	if(reply)
	{
		//this will not work if there are multiple processes sending request at same time.
		if(reply->type == SAVEUSER_REPLY)
		{
                        if(strlen(reply->data.userRecord.userName) > 0)
                        {
				result = TRUE;
			}
		}
		free(reply);
	}

	return result;
}


//************************************************************************************
// This function is called by the lower privilege process to open the iptables
// firewall.
//
//************************************************************************************

BOOL psOpenFirewall(DWORD trustedIP, WORD allocPort)
{
	PRIVPROC_SUPPORT pps;

	pps.type = OPEN_FW;
	pps.data.fwParams.allowedIP = trustedIP;
	pps.data.fwParams.portNum = allocPort;

	ssize_t wrote = write(gPrivProcCommandPipe[1], &pps, sizeof(pps));
	if(wrote == sizeof(pps))
		return TRUE;
	else
		return FALSE;
}


//************************************************************************************
// This function is called by the lower privilege thread to close the iptables
// firewall.
//
//************************************************************************************

BOOL psCloseFirewall(DWORD trustedIP, WORD allocPort)
{
	PRIVPROC_SUPPORT pps;

	pps.type = CLOSE_FW;
	pps.data.fwParams.allowedIP = trustedIP;
	pps.data.fwParams.portNum = allocPort;

	ssize_t wrote = write(gPrivProcCommandPipe[1], &pps, sizeof(pps));
	if(wrote == sizeof(pps))
		return TRUE;
	else
		return FALSE;
}


BOOL auxSendUDPReply(DWORD destIP, DWORD destPort, DWORD srcPort, char *reply, DWORD pktSize)
{
	SOCKET sock;
	//****************************** Send PortSeq Request PKT ****************************/
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if(sock == INVALID_SOCKET)
	{
		logString("Error: UDP socket creation failed: %d\n", errno);
		return FALSE;
	}
        struct sockaddr_in dest;
        dest.sin_family = AF_INET;
        dest.sin_port = htons(destPort);
        dest.sin_addr.s_addr = destIP;

        struct sockaddr_in local;
        local.sin_family = AF_INET;
        local.sin_port = htons(srcPort);
        local.sin_addr.s_addr = htonl(INADDR_ANY);
	
	if(bind(sock, (struct sockaddr *)&local, sizeof(local)) == SOCKET_ERROR)
	{
		logString("Error: UDP socket bind() failed!\n");
		close(sock);
		return FALSE;
	}
	
	if(connect(sock, (struct sockaddr *)&dest, sizeof(dest)) == SOCKET_ERROR)
	{
		logString("Error: UDP connect() failed!\n");
		close(sock);
		return FALSE;
	}
                
	if(send(sock, reply, pktSize, 0) == SOCKET_ERROR)
	{
		logString("Error: UDP send() failed!\n");
		close(sock);
		return FALSE;
	}

        close(sock);
        return TRUE;
}


//************************************************************************************
// This function is to be used by the lower priv child to send UDP reply P3 back to
// the client.  This is required since we may need to bind to a privileged port.
//
//************************************************************************************

BOOL psSendUDPReply(DWORD destIP, WORD destPort, WORD srcPort, char *reply, DWORD pktSize)
{
	PRIVPROC_SUPPORT pps;
	PRIVPROC_SUPPORT *ppsReply;
	BOOL result = FALSE;

	pps.type = SEND_UDPDATA;
	pps.data.udpSendParams.destIP = destIP;
	pps.data.udpSendParams.destPort = destPort;
	pps.data.udpSendParams.localPort = srcPort;
	if(pktSize > MAX_UDP_DATA)
		pktSize = MAX_UDP_DATA;
	pps.data.udpSendParams.pktSize = pktSize;

	memcpy(pps.data.udpSendParams.data, reply, pktSize);

	ssize_t wrote = write(gPrivProcCommandPipe[1], &pps, sizeof(pps));
	if(wrote != sizeof(pps))
		return FALSE;

        ppsReply = (PRIVPROC_SUPPORT *)psGetPipeReply(sizeof(PRIVPROC_SUPPORT));
        if(ppsReply)
	{
		if(ppsReply->type == SEND_UDPDATA_OK)
		{
			result = TRUE;
		}
		free(ppsReply);
	}
	return result;				
}


//************************************************************************************
// This thread function receives request to perform certain tasks from a lower
// privileged process.
//
//************************************************************************************

LPVOID privProcThreadFunc(LPVOID data)
{
	close(gPrivProcCommandPipe[1]);
	close(gPrivProcReplyPipe[0]);
	
	while(true)
	{
		PRIVPROC_SUPPORT pps;
		PRIVPROC_SUPPORT ppsReply;
		ssize_t recv = 0;

		while(recv < sizeof(PRIVPROC_SUPPORT))
		{
			ssize_t curRecv = read(gPrivProcCommandPipe[0], 
					       (LPVOID)( ((DWORD)&pps) + recv ), 
					       sizeof(PRIVPROC_SUPPORT) - recv);
			if(curRecv <= 0)
				goto done;
			else
				recv += curRecv;
		}
		
		if(recv == sizeof(PRIVPROC_SUPPORT))
		{
			if(pps.type == CLOSE_FW)
			{
				//printf("CloseFW\n");
				fwCloseFirewall(pps.data.fwParams.allowedIP, 
						pps.data.fwParams.portNum);
			}
			else if(pps.type == OPEN_FW)
			{
				//printf("OpenFw %d\n", pps.data.fwParams.portNum);
				fwOpenFirewall(pps.data.fwParams.allowedIP, 
						pps.data.fwParams.portNum);
			}
			else if(pps.type == GETUSER)
			{
				USER_RECORD userRec;
				if(!getUserRecord(USER_FILENAME, pps.data.userRecord.userName, &userRec))
				{
					//printf("PrivProc -> (%s) getUserRecord bad\n", pps.data.userRecord.userName);
					memset(&userRec, 0, sizeof(USER_RECORD));
				}
				ppsReply.type = GETUSER_REPLY;
				memcpy(&(ppsReply.data.userRecord), &userRec, sizeof(USER_RECORD)); 
				write(gPrivProcReplyPipe[1], &ppsReply, sizeof(PRIVPROC_SUPPORT));	
			}
			else if(pps.type == SAVEUSER)
			{
				memset(&ppsReply, 0, sizeof(ppsReply));
				if(updateUserTimestamp(USER_FILENAME, pps.data.userRecord.userName, 
						    	pps.data.userRecord.lastTime))
				{
					ppsReply.data.userRecord = pps.data.userRecord;
				}
				ppsReply.type = SAVEUSER_REPLY;
				write(gPrivProcReplyPipe[1], &ppsReply, sizeof(PRIVPROC_SUPPORT));
			}
			else if(pps.type == SEND_UDPDATA)
			{
				memset(&ppsReply, 0, sizeof(ppsReply));
				if(auxSendUDPReply(pps.data.udpSendParams.destIP,
						pps.data.udpSendParams.destPort,
						pps.data.udpSendParams.localPort,
						(char *)pps.data.udpSendParams.data,
						pps.data.udpSendParams.pktSize))
				{
					ppsReply.type = SEND_UDPDATA_OK;
				}
				else
					ppsReply.type = SEND_UDPDATA_NOTOK;
				write(gPrivProcReplyPipe[1], &ppsReply, sizeof(PRIVPROC_SUPPORT));
			}
			else if(pps.type == WRITE_SYSLOG)
			{
				pps.data.syslogParams.logStr[MAX_LOG_LEN] = 0;
				syslog(LOG_MAKEPRI(LOG_DAEMON, LOG_INFO), pps.data.syslogParams.logStr);
			}
		}
	}
done:
	return NULL;
}


