//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

//************************************************************************************
// This files contains functions that performs logging for the server.
//
//************************************************************************************

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <syslog.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include "windowsdef.h"
#include "semaphore.h"
#include "privProcSupport.h"


FILE 		*logFile = NULL;
sem_t 		gLogFileCritSect;

extern BOOL 	gUseSyslog;
extern int 	gPrivProcCommandPipe[2];


void initLog(char *logFilename)
{
	if(gUseSyslog)
	{
		openlog("sig2knockd", 0, LOG_USER);
	}
	else
	{
		if(logFilename)
		{	
			logFile = fopen(logFilename, "a");
		}
		else
		{
			logFile = stdout;
		}
	}
}


void logString(char *str, ...)
{
	if(logFile || gUseSyslog)
	{
		struct tm st;
		time_t ttCurTime;

		time(&ttCurTime);
		localtime_r(&ttCurTime, &st);

		va_list p;
		va_start(p, str);

		if(gUseSyslog)
		{
			if(geteuid() == 0)	// root
			{
				// check format string vuln
				vsyslog(LOG_MAKEPRI(LOG_DAEMON, LOG_INFO), str, p);
			}
			else
			{
				PRIVPROC_SUPPORT request;
				memset(&request, 0, sizeof(request));
				request.type = WRITE_SYSLOG;

				vsnprintf(request.data.syslogParams.logStr, MAX_LOG_LEN, str, p);
				write(gPrivProcCommandPipe[1], &request, sizeof(request));
			}
		}
		else
		{		
			fprintf(logFile, "%.4d-%.2d-%.2d %.2d:%.2d.%.2d ",
				st.tm_year + 1900, st.tm_mon + 1, st.tm_mday,
				st.tm_hour, st.tm_min, st.tm_sec);

			vfprintf(logFile, str, p);
			fflush(logFile);
		}
		va_end(p);
	}
}


void endLog(void)
{
	if(logFile)
	{
		fclose(logFile);
		logFile = NULL;
	}
	if(gUseSyslog)
		closelog();
}
