//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************


//************************************************************************************
// Prints a list of pcap interfaces and usage instructions
//
//
//************************************************************************************

void printUsage(char *exeName)
{
	pcap_if_t *alldevs;
    pcap_if_t *d;
    int interfaceCount;
    char errbuf[PCAP_ERRBUF_SIZE];
    
    // Retrieve the device list
    if (pcap_findalldevs(&alldevs, errbuf) == -1)
    {
        printf("Error  in pcap_findalldevs: %s\n", errbuf);
        return;
    }
    
	interfaceCount = 0;

    // Print the list
    for(d = alldevs; d; d = d->next)
    {		
		for(pcap_addr_t *a = d->addresses; a; a = a->next) 
		{
			if(a->addr->sa_family == AF_INET)
			{
				if (d->description)
					printf("%2d - (%s)\n", interfaceCount, d->description);
				else            
					printf("%2d - (No description available)\n", interfaceCount);
				printf("     %s\n\n", d->name);
			}
		}
		interfaceCount++;
    }
    
    if(interfaceCount == 0)
    {
        printf("\nNo interfaces found! Make sure WinPcap is installed.\n");
        return;

    }
	printf("\nUsage: %s -i <interface>", exeName);
	printf("\n       %s -i <interface> -s     (install as service)", exeName);
	printf("\n       %s -u                    (uninstall service)\n", exeName);
	printf("\nExample: %s -i 2      (to run on console and sniff on interface 2)", exeName);
	printf("\n         %s -i 2 -s   (install as service using interface 2)\n", exeName);
	
	printf("\nWhen running as a service, all log entries will go to %s\n\n", SERVER_LOGFILE);

    // We don't need any more the device list. Free it
    pcap_freealldevs(alldevs);
}



//************************************************************************************
// Thread that deletes stale port sequence request records.
//
//************************************************************************************

DWORD WINAPI stalePortReqCleanupThreadFunc(LPVOID data)
{
	while(true && !gStopFlag)
	{
		EnterCriticalSection(&gPortReqRequestListControl);
		
		deleteMultipleItems(&gPortSeqRequestList, deleteStalePortSeqRecFunc, NULL);
		
		//printList(&gPortSeqRequestList, printPortSeqReqRec);

		LeaveCriticalSection(&gPortReqRequestListControl);

		Sleep(2000);
	}
	return 0;
}






//************************************************************************************
// ServiceHandler
//
//************************************************************************************

void ServiceHandler(DWORD dwCommand)
{
	switch(dwCommand)
	{
		case SERVICE_CONTROL_STOP:
		{
			updateOurStatus(SERVICE_STOP_PENDING, 0, 1);

			pcap_close(gPcapHandle);
			gStopFlag = true;
			
			Sleep(1000);
			logString("Server Terminated.\n"); 
			endLog();
			
			updateOurStatus(SERVICE_STOPPED, 0, 0);
			break;
		}
		default:
			break;
	}
}


//************************************************************************************
// ServiceMain
//
//************************************************************************************

VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv)
{
	gSvcHandle = RegisterServiceCtrlHandler(OUR_SVCNAME, (LPHANDLER_FUNCTION)&ServiceHandler);
	gStopFlag = false;

	if(gSvcHandle == 0)
		return;

	updateOurStatus(SERVICE_START_PENDING, 0, 1);

	initList(&gPortSeqRequestList);
	initLog(SERVER_LOGFILE);		// log to stdout	
	InitializeCriticalSection(&gPortReqRequestListControl);
	
	logString("Server started.\n");

	updateOurStatus(SERVICE_RUNNING, 0, 0);
				
	// start the capture
	pcap_loop(gPcapHandle, 0, packet_handler, NULL);

}


//************************************************************************************
// Opens winpcap and returns its handle
//
//************************************************************************************

pcap_t *openPcapInterface(short interfaceNum)
{
	pcap_t *hPcap = NULL;
	pcap_if_t *alldevs;
    pcap_if_t *d, *toOpen = NULL; 
    char errbuf[PCAP_ERRBUF_SIZE];
    
    // Retrieve the device list
    if (pcap_findalldevs(&alldevs, errbuf) == -1)
    {
        printf("Error in pcap_findalldevs: %s\n", errbuf);
        return NULL;
    }
    
    // Print the list
	int count = 0;
    for(d = alldevs; d; d = d->next)
    {		
		if(count == interfaceNum)
		{
			toOpen = d;
			pcap_addr_t *a = d->addresses;
			
			gOurIPAddr = ((struct sockaddr_in *)a->addr)->sin_addr;
        
			strcpy(gOurIPAddrStr, inet_ntoa(gOurIPAddr));

			printf("IP Address = %s\n", gOurIPAddrStr);
			printf("\n");
			printConfig();
			printf("\n");

			break;
		}
		count++;
    }
    
    if(toOpen == NULL)
    {
        printf("\nInterface not found!  Make sure WinPcap is installed.\n");
		if(alldevs)
			pcap_freealldevs(alldevs);
        return NULL;
    }

	printf("Opening %s\n\n", toOpen->name);
	
    if ( (hPcap = pcap_open_live(toOpen->name,		// name of the device
                             1000,						// portion of the packet to capture. 
                             FALSE,						// promiscuous mode
                             500,						// read timeout
                             errbuf						// error buffer
                             ) ) == NULL)
	
    {
        printf("\nUnable to open the adapter. Make sure you are running this as Administrator\n");        
		// Free the device list
		pcap_freealldevs(alldevs);
		return NULL;
	}

    // Free the device list
	pcap_freealldevs(alldevs);

	return hPcap;
}


//************************************************************************************
// Change the current directory to the one where the EXE, configuration file and user
// acct file are.  This is required if we are running as service, since the default 
// directory will be \windows\system32.
//
//************************************************************************************

void setWorkingDirectory(void)
{
	char exePath[MAX_PATH+2];
	memset(exePath, 0, sizeof(exePath));
	if(GetModuleFileName(NULL, exePath, sizeof(exePath)-1) > 0)
	{
		char *ptr = exePath + strlen(exePath) - 1;
		while(*ptr && ptr > exePath && *ptr != '\\')
			ptr--;

		if(*ptr == '\\')
			*ptr = 0;
		SetCurrentDirectory(exePath);
	}
}


//************************************************************************************
// Program entry
//
//************************************************************************************

int main(int argc, char* argv[])
{
	printf("DrKnockd Version %s Copyright (c) 2006 Olivier ROCH VILATO and Chew Keong TAN\n\n", PROG_VERSION);

	if(argc < 2)
	{
		printUsage(argv[0]);
		return 1;
	}

	setWorkingDirectory();
	FILE *fileChk = fopen(USER_FILENAME, "r");
	if(fileChk == NULL)
	{
		printf("You must first create a user account file (%s), and use\n"
		       "drknockd_useradd to create accounts.\n\n", USER_FILENAME);
		return 1;
	}
	fclose(fileChk);

	if(!initRand())
	{
		printf("Error initializing random generator!\n");
		return 1;
	}

	if(!loadServerConfig(CONFIG_FILENAME))
	{
		printf("Configuration file <%s> not found! Using defaults.\n", CONFIG_FILENAME);
	}

	int opt;
	char *interfaceNum = NULL;
	BOOL optInstallSvc = FALSE;
	BOOL optUninstallSvc = FALSE;

	while((opt = getopt(argc, argv, "i:su")) != -1)
	{
		switch(opt)
		{
			case 'u':				
				if(optInstallSvc)
				{
					printf("Conflicting arguments. Install and uninstall options cannot be used together\n");
					return 1;
				}
				optUninstallSvc = TRUE;
				break;
			case 'i':				
				interfaceNum = optarg;
				break;
			case 's':
				if(optUninstallSvc)
				{
					printf("Conflicting arguments. Install and uninstall options cannot be used together\n");
					return 1;
				}
				optInstallSvc = TRUE;
				break;
			default:
				break;
		}
	}
	
	if(interfaceNum)
	{
		if(optUninstallSvc)
		{
			printUsage(argv[0]);
			return 1;			
		}

		short int nInterface = atoi(interfaceNum);		
	
		gPcapHandle = openPcapInterface(nInterface);
		if(gPcapHandle == NULL)
			return 1;

		if(optInstallSvc)
		{
			pcap_close(gPcapHandle);

			char exeArgs[21];
			_snprintf(exeArgs, sizeof(exeArgs) - 1, " -i %d", nInterface);

			installSvc(OUR_SVCNAME, "DrKnockd Win32 Server", exeArgs);
		}
		else
		{
			DWORD svcStatus = getServiceStatus(OUR_SVCNAME);

			if(svcStatus == SERVICE_START_PENDING)
			{
				SERVICE_TABLE_ENTRY stb1[] = 
				{
					{OUR_SVCNAME, (LPSERVICE_MAIN_FUNCTION)&ServiceMain },
					{NULL, NULL}
				};

				StartServiceCtrlDispatcher(stb1);
			}
			else		// run on console
			{
				initList(&gPortSeqRequestList);
				initLog(NULL);		// log to stdout	
				InitializeCriticalSection(&gPortReqRequestListControl);
					
				logString("Server started.\n");

				// start the capture
				pcap_loop(gPcapHandle, 0, packet_handler, NULL);

				logString("Server Terminated.\n"); 
				endLog();
			}
		}
	}
	else
	{
		if(optUninstallSvc)
			deleteSvc(OUR_SVCNAME);
		else
		{
			printUsage(argv[0]);
			return 1;
		}
	}

	return 0;
}


