//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#include <winsock2.h>
#include <windows.h>
#include <time.h>
#include "portForward.h"
#include "randomGen.h"
#include "serverLog.h"
#include "serverConfig.h"


//************************************************************************************
// Executes pktctl.exe to open firewall for accepted clients
//
//************************************************************************************

BOOL openFirewall(DWORD trustedIP, WORD allocPort, const char *pktFilterInterface)
{
	if(!pktFilterInterface || strlen(pktFilterInterface) == 0 ||
		strlen(pktFilterInterface) > 5)
		return FALSE;

	char command[257];
	struct in_addr ina;
	ina.S_un.S_addr = trustedIP;

	_snprintf(command, sizeof(command) - 1,
		      "pktctl -a \"pass in on %s proto tcp from %s to any port = %u\"",
			  pktFilterInterface, inet_ntoa(ina), allocPort);

	STARTUPINFO sinfo;
	PROCESS_INFORMATION pinfo;

	memset(&sinfo, 0, sizeof(sinfo));
	memset(&pinfo, 0, sizeof(pinfo));

	sinfo.cb = sizeof(sinfo);
	
	if(CreateProcess(NULL, command, NULL, NULL, FALSE, 0, NULL, NULL, &sinfo, &pinfo))
	{
		return TRUE;
	}
	return FALSE;
}


//************************************************************************************
// Executes pktctl.exe to close firewall after client has disconnected
//
//************************************************************************************

BOOL closeFirewall(DWORD trustedIP, WORD allocPort, const char *pktFilterInterface)
{
	if(!pktFilterInterface || strlen(pktFilterInterface) == 0 ||
		strlen(pktFilterInterface) > 5)
		return FALSE;

	BOOL result = FALSE;
	STARTUPINFO sinfo;
	PROCESS_INFORMATION pinfo;

	memset(&sinfo, 0, sizeof(sinfo));
	memset(&pinfo, 0, sizeof(pinfo));

	HANDLE hReadPipe1, hWritePipe1; 
	HANDLE hReadPipe2, hWritePipe2;

	SECURITY_ATTRIBUTES sa; 
	sa.nLength = sizeof(sa); 
	sa.lpSecurityDescriptor = 0; 
	sa.bInheritHandle = TRUE; 

	if(CreatePipe(&hReadPipe1, &hWritePipe1, &sa, 0) == 0)
	{
		//printf("Bad CreatePipe()\n");
		return FALSE;
	}
	if(CreatePipe(&hReadPipe2, &hWritePipe2, &sa, 0) == 0)
	{
		//printf("Bad CreatePipe()\n");
		CloseHandle(hReadPipe1);
		CloseHandle(hWritePipe1);
		return FALSE;
	}
	
	// link pipes to the stdin and stdout of the processes

	sinfo.cb = sizeof(STARTUPINFO);
	sinfo.dwFlags = STARTF_USESTDHANDLES; 
	sinfo.hStdInput = hReadPipe1;            
	sinfo.hStdOutput = sinfo.hStdError = hWritePipe2;

	char command[257];
	_snprintf(command, sizeof(command) - 1, "pktctl -L %s", pktFilterInterface);
	char strToSearch[20];
	_snprintf(strToSearch, sizeof(strToSearch) - 1, "port = %u", allocPort);
	int ruleNum = 0;

	//printf("command = <%s>\n", command);
	//printf("search = %s\n", strToSearch);

	DWORD bufIncSize = 65536;
	DWORD curSize = 0;	
	DWORD sizeLeft = bufIncSize;

	char *outputBuffer = (char *)malloc(bufIncSize);
	if(!outputBuffer)
		goto done;


	if(CreateProcess(NULL, command, NULL, NULL, TRUE, 0, NULL, NULL, &sinfo, &pinfo))
	{
		int count = 0;
		while(true)
		{			
			DWORD bytesRead = 0;
			
			int readResult = PeekNamedPipe(hReadPipe2, NULL, 0, NULL, &bytesRead, NULL);
			if(readResult == 0)
				break;
			
			if(bytesRead != 0)
			{
				count = 0;
				readResult = ReadFile(hReadPipe2, &outputBuffer[curSize], 
					                  sizeLeft, &bytesRead, 0);

				curSize += bytesRead;
				sizeLeft -= bytesRead;
				if(sizeLeft == 0)
				{
					char *newBuffer = (char *)malloc(curSize + bufIncSize);
					if(newBuffer)
					{
						memcpy(newBuffer, outputBuffer, curSize);
						free(outputBuffer);
						outputBuffer = newBuffer;
						sizeLeft = bufIncSize;
					}
					else
						break;
				}
			}
			else
			{
				if(count < 3)
				{
					count++;
					Sleep(500);
				}
				else
					break;
			}
		}

		for(DWORD i = 0; i < curSize - strlen(strToSearch); i++)
		{
			if(memcmp(&outputBuffer[i], strToSearch, strlen(strToSearch)) == 0)
			{
				char *ptr = &outputBuffer[i];
				while(ptr > outputBuffer && *ptr != ':')
					ptr--;

				if(*ptr == ':')
				{
					*ptr = 0;
					ptr--;
					while(ptr > outputBuffer && (*ptr >= 0 && *ptr <= 9))
						ptr--;
					ruleNum = atoi(ptr);
				}
				break;
			}
		}
		free(outputBuffer);

		if(ruleNum)
		{			
			_snprintf(command, sizeof(command) - 1, "pktctl -d %d %s", ruleNum,
					  pktFilterInterface);
			//printf("command = <%s>\n", command);

			memset(&sinfo, 0, sizeof(sinfo));
			memset(&pinfo, 0, sizeof(pinfo));

			sinfo.cb = sizeof(sinfo);
	
			if(CreateProcess(NULL, command, NULL, NULL, FALSE, 0, NULL, NULL, &sinfo, &pinfo))
				result = TRUE;
		}
	}

done:
	CloseHandle(hReadPipe1);
	CloseHandle(hReadPipe2);
	CloseHandle(hWritePipe1);
	CloseHandle(hWritePipe2);
	return result;
}





//************************************************************************************
// Blocking send
//
//************************************************************************************

BOOL sendSockData(SOCKET sock, char *buffer, int size)
{
	BOOL result = FALSE;
	while(TRUE)
	{
		if(send(sock, buffer, size, 0) == SOCKET_ERROR)
		{
			int status = WSAGetLastError();
			if(status == WSAEWOULDBLOCK)
			{					
				HANDLE hSockEvent = WSACreateEvent();					
				WSAEventSelect(sock, hSockEvent, FD_WRITE | FD_CLOSE);
					
				WaitForSingleObject(hSockEvent, INFINITE);
		
				WSANETWORKEVENTS wsaEvt;

				WSAEnumNetworkEvents(sock, hSockEvent, &wsaEvt);		// this will set the event to non-signaled state
				WSACloseEvent(hSockEvent);

				if(wsaEvt.lNetworkEvents & FD_CLOSE)
				{									
					break;
				}
			}
			else
				break;
		}
		else
		{
			result = TRUE;
			break;
		}
	}

	return result;
}


//************************************************************************************
// Function to perform port forwarding. 
//
//************************************************************************************

DWORD forwardFunc(SOCKET client, struct in_addr ina, WORD portNum)
{	
	char buffer[65536];

	SOCKET f_sock = WSASocket(AF_INET, SOCK_STREAM, 0, NULL, 0, 0);
	if(f_sock == INVALID_SOCKET)
		goto done;

	struct sockaddr_in sin;
	sin.sin_family = AF_INET;
	sin.sin_port = htons(getConfigForwardToPort());
	sin.sin_addr.s_addr = getConfigForwardToIP();

	if(connect(f_sock, (struct sockaddr *)&sin, sizeof(sin)) == SOCKET_ERROR)
	{
		logString("Connection to forwarded IP failed!\n");
		closesocket(f_sock);		
	}
	else
	{
		HANDLE hSockEvent1 = WSACreateEvent();
		HANDLE hSockEvent2 = WSACreateEvent();

		while(true)
		{
			WSAEventSelect(client, hSockEvent1, FD_READ | FD_CLOSE);
			WSAEventSelect(f_sock, hSockEvent2, FD_READ | FD_CLOSE);

			HANDLE events[2];
			events[0] = hSockEvent1;
			events[1] = hSockEvent2;

			int iEvent = WaitForMultipleObjects(2, events, FALSE, INFINITE);

			if(iEvent == 0)		// read event from "client" socket 
			{
				WSANETWORKEVENTS wsaEvt;

				WSAEnumNetworkEvents(client, hSockEvent1, &wsaEvt);		// this will set the event to non-signaled state
				if(wsaEvt.lNetworkEvents & FD_READ)
				{					
					int lBytesRead = recv(client, buffer, sizeof(buffer), 0);

					if(lBytesRead > 0)
					{
						sendSockData(f_sock, buffer, lBytesRead);
					}
					else
						break;
				}
				if(wsaEvt.lNetworkEvents & FD_CLOSE)
				{
					break;
				}
			}
			else		// read/close event from f_sock socket
			{
				WSANETWORKEVENTS wsaEvt;

				WSAEnumNetworkEvents(f_sock, hSockEvent2, &wsaEvt);		// this will set the event to non-signaled state
				if(wsaEvt.lNetworkEvents & FD_READ)
				{
					int lBytesRead = recv(f_sock, buffer, sizeof(buffer), 0);

					if(lBytesRead > 0)
					{
						sendSockData(client, buffer, lBytesRead);
					}
					else
						break;
				}
				if(wsaEvt.lNetworkEvents & FD_CLOSE)
				{
					break;
				}
			}
		}
		WSACloseEvent(hSockEvent1);
		WSACloseEvent(hSockEvent2);
		closesocket(f_sock);
	}

done:
	logString("Client %s disconnected from port %d.\n", inet_ntoa(ina), portNum);
	closesocket(client);
	
	return 0;
}


//************************************************************************************
// Thread function to serve accepted clients.  In this version, we support only
// forwarding of one connection per request. i.e. The allocated port will only close
// after the first connection from the client is served.
//
//************************************************************************************

DWORD WINAPI forwarderThreadFunc(LPVOID data)
{
	FORWARDER_DATA *fdata = (FORWARDER_DATA *)data;
	SOCKET s = fdata->s;
	DWORD allowedIP = fdata->allowedIP;
	WORD portNum = fdata->portNum;

	free(data);
	struct sockaddr_in sin;
	HANDLE hAcceptEvent = WSACreateEvent();					
	WSAEventSelect(s, hAcceptEvent, FD_ACCEPT);

	DWORD portOpenTime = getConfigSingleConnPortOpenTime();		// in secs
	bool allowMultipleConnections=getConfigAllowMultipleConnections();
	DWORD startTime = time(NULL);
	DWORD timeLeft;

	openFirewall(allowedIP, portNum, getPktFilterInterface());		// open firewall

	while(true)
	{
		timeLeft = portOpenTime - (time(NULL) - startTime);
		
		if(WaitForSingleObject(hAcceptEvent, timeLeft * 1000) == WAIT_TIMEOUT)
		{
			struct in_addr ina1;
			ina1.S_un.S_addr = allowedIP;

			logString("Port %d for %s timeout and closed.\n", portNum, inet_ntoa(ina1));
			goto done;
		}		

		WSANETWORKEVENTS wsaEvt;
		WSAEnumNetworkEvents(s, hAcceptEvent, &wsaEvt);		// this will set the event to non-signaled state

		if(wsaEvt.lNetworkEvents & FD_ACCEPT)
		{
			int sin_size = sizeof(sin);
			SOCKET client = WSAAccept(s, (SOCKADDR *)&sin, &sin_size, NULL, 0);

			if(client != INVALID_SOCKET)
			{
				if(sin.sin_addr.S_un.S_addr != allowedIP)
				{
					// unauthorized connection
					logString("Unauthorized connection to port %d from %s!\n", 
						       portNum, inet_ntoa(sin.sin_addr));

					closesocket(client);
					continue;
				}
				else
				{
					logString("Client from %s connected to port %d.\n", 
						      inet_ntoa(sin.sin_addr), portNum);
					forwardFunc(client, sin.sin_addr, portNum);
				}
				if (!allowMultipleConnections) { goto done; } 

			}
			else
			{
				logString("Error: Socket is invalid in forwarder thread.\n");
			}
		}
	}

done:
	closeFirewall(allowedIP, portNum, getPktFilterInterface());		// open firewall
	WSACloseEvent(hAcceptEvent);
	closesocket(s);
	
	return 0;
}
