//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

//************************************************************************************
// This files contains functions that performs logging for the server.
//
//************************************************************************************

#include <stdio.h>
#include <stdarg.h>
#include <windows.h>


FILE *logFile = NULL;
CRITICAL_SECTION gLogFileCritSect;


void initLog(char *logFilename)
{
	InitializeCriticalSection(&gLogFileCritSect);
	if(logFilename)
	{
		logFile = fopen(logFilename, "a");
	}
	else
		logFile = stdout;
}


void logString(char *str, ...)
{
	if(logFile)
	{
		SYSTEMTIME st;
		GetLocalTime(&st);
		va_list p;
		va_start(p, str);

		EnterCriticalSection(&gLogFileCritSect);

		fprintf(logFile, "%.4d-%.2d-%.2d %.2d:%.2d.%.2d ", st.wYear, st.wMonth, st.wDay,
				st.wHour, st.wMinute, st.wSecond);
		vfprintf(logFile, str, p);		
		fflush(logFile);
		
		LeaveCriticalSection(&gLogFileCritSect);
	}
}


void endLog(void)
{
	if(logFile)
	{
		EnterCriticalSection(&gLogFileCritSect);
		fclose(logFile);
		logFile = NULL;
		LeaveCriticalSection(&gLogFileCritSect);
	}
	DeleteCriticalSection(&gLogFileCritSect);
}
