//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#include <windows.h>
#include "userRecord.h"
#include "base64.h"
#include "randomGen.h"


//****************************************************************************************
// User account file is in the following format
//
// #userid:password hash:timestamp
//
// mary:OpqhDisuS7jGQeeNe7tUIQ==:1083037436
// peter:arTN/spZ3MXPN2AhqH2Akg==:1082993487
//
//****************************************************************************************

void strTrim(char *str)
{
	int pos = strlen(str) - 1;
	
	while(pos >= 0)
	{
		if(str[pos] == ' ' || str[pos] == '\n' || 
		   str[pos] == '\r' || str[pos] == '\t')
		{
			str[pos] = 0;
		}
		else
		{
			//printf("==> %.2X\n", str[pos]);
			break;
		}		
		pos--;
	}

	char *ptr = str;
	while(*ptr == ' ' || *ptr == '\t')
		ptr++;

	if(ptr != str)
	{
		while(*ptr)
		{
			*str++ = *ptr++;
		}
		*str = 0;
	}
}


int countTokenChar(char *ptr, char tokenChar)
{
	int tokens = 0;
	while(*ptr)
	{
		if(*ptr++ == tokenChar)
			tokens++;
	}

	return tokens;
}


//************************************************************************************
// Retrieves the specified user record from user.txt file
//
//************************************************************************************

BOOL getUserRecord(char *filename, char *userName, USER_RECORD *userRec)
{
	FILE *fp;

	fp = fopen(filename, "r");
	if(fp)
	{
		char buffer[257];
		while(fgets(buffer, sizeof(buffer), fp))
		{
			strTrim(buffer);
			if(strlen(buffer) && buffer[0] != '#')		// comment char = #
			{
				if(countTokenChar(buffer, ':') == 2)
				{					
					char *ptr = buffer;
					while(*ptr != ':')
						ptr++;

					*ptr = 0;
					if(strcmp(buffer, userName) == 0)
					{
						memset(userRec, 0, sizeof(USER_RECORD));
                        int i;
						for(i = 0; i < sizeof(userRec->userName) - 1; i++)
							userRec->userName[i] = buffer[i];

						char *ptr2 = ptr;
						while(*ptr2 != ':')
							ptr2++;

						*ptr2 = 0;
						ptr++;

						unsigned char passwordHashB64[SIZEOF_USERPASSWORD_HASHB64];

						for(i = 0; i < sizeof(passwordHashB64); i++)
							passwordHashB64[i] = ptr[i];

						B64Decode(passwordHashB64, SIZEOF_USERPASSWORD_HASHB64, 
									userRec->userPasswordHash);

						userRec->lastTime = atol(ptr2 + 1);		
						
						fclose(fp);
						return TRUE;
					}
				}
			}
		}
		fclose(fp);
	}

	return FALSE;
}

char safeCharSet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

//************************************************************************************
// Updates timestamp in user record file
//
//************************************************************************************

BOOL updateUserTimestamp(char *filename, char *userName, DWORD newTimestamp)
{
	FILE *fp;

	fp = fopen(filename, "r");
	if(fp)
	{
		FILE *fpout;
		char tempFile[MAX_PATH + 1];

		char randPrefix[4];
		getRandBytes((unsigned char *)randPrefix, sizeof(randPrefix) - 1);

		for(int i = 0; i < sizeof(randPrefix) - 1; i++)
			randPrefix[i] = safeCharSet[randPrefix[i] % strlen(safeCharSet)];
		randPrefix[sizeof(randPrefix) - 1] = 0;

		GetTempFileName(".", randPrefix, 0, tempFile);		

		fpout = fopen(tempFile, "w");
		if(!fpout)
		{
			fclose(fp);
			goto done;
		}

		char buffer[257];
		while(fgets(buffer, sizeof(buffer), fp))
		{
			strTrim(buffer);
			if(strlen(buffer) && buffer[0] != '#')		// comment char = #
			{
				if(countTokenChar(buffer, ':') == 2)
				{					
					char *ptr = buffer;
					while(*ptr != ':')
						ptr++;

					*ptr = 0;
					if(strcmp(buffer, userName) == 0)
					{
						USER_RECORD userRec;
						memset(&userRec, 0, sizeof(USER_RECORD));
						int i;
						for(i = 0; i < sizeof(userRec.userName) - 1; i++)
							userRec.userName[i] = buffer[i];

						char *ptr2 = ptr;
						while(*ptr2 != ':')
							ptr2++;

						*ptr2 = 0;
						ptr++;

						unsigned char passwordHashB64[SIZEOF_USERPASSWORD_HASHB64 + 1];
						memset(passwordHashB64, 0, SIZEOF_USERPASSWORD_HASHB64 + 1);

						for(i = 0; i < sizeof(passwordHashB64); i++)
							passwordHashB64[i] = ptr[i];

						userRec.lastTime = atol(ptr2 + 1);

						fprintf(fpout, "%s:%s:%u\n", userRec.userName, passwordHashB64, newTimestamp);
					}
					else
					{
						*ptr = ':';
						fprintf(fpout, "%s\n", buffer);
					}
				}
			}
			else
			{
				fprintf(fpout, "%s\n", buffer);
			}
		}
		fclose(fpout);
		fclose(fp);
		CopyFile(tempFile, filename, FALSE);
		DeleteFile(tempFile);
		return TRUE;
	}
done:
	return FALSE;
}
