//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <windows.h>
#include <stdio.h>
#include "win32Service.h"


extern SERVICE_STATUS_HANDLE	gSvcHandle;


//************************************************************************************
// Install this program as a service.
//
//************************************************************************************

char installSvc(char *srvName, char *srvFullName, char *exeArgs)
{
	SC_HANDLE scm, srv;
	DWORD status;
	char exeName[MAX_PATH + 1];
	
	scm = OpenSCManager(
			NULL,						// local machine
			NULL,						// SERVICE_ACTIVE_DATABASE
			SC_MANAGER_CREATE_SERVICE);
	
	if (scm == NULL) 
	{
		printf("Error opening SC Manager!  Installation of service failed.\n");
		return 1;
	}
	
	srv = OpenService(scm, srvName, GENERIC_READ);
	if(srv != NULL)
	{
		printf("Service is already installed!\n");
		CloseServiceHandle(srv);
		CloseServiceHandle(scm);
		return 0;
	}

	memset(exeName, 0, sizeof(exeName));
	status = GetModuleFileName(NULL, exeName, sizeof(exeName)-1);
	
	if (!status) 
	{
		printf("Error determining path of executable!  Installation failed.\n");
		return 1;
	}
	strncat(exeName, exeArgs, MAX_PATH);

	// Create the service
	srv = CreateService(
		scm,							// handle to SCM
		srvName,						// name of Win32 service
		srvFullName,					// full name
		SERVICE_ALL_ACCESS,				// all access
		SERVICE_WIN32_OWN_PROCESS,		// a service that runs in its own process
		SERVICE_AUTO_START,				// SERVICE_DEMAND_START launched at startup
		SERVICE_ERROR_IGNORE,			// start error will be ignored
		exeName, 
		NULL,							// no load order group
		NULL,							// no tag
		"NPF\x00\x00",					// we need winpcap driver to be started first
		NULL,							// LocalSystem account
		NULL							// LocalSystem
		);

	if (srv != NULL)
	{
		SERVICE_DESCRIPTION sd;
		sd.lpDescription = srvFullName;

		ChangeServiceConfig2(srv, SERVICE_CONFIG_DESCRIPTION, &sd);

		printf("Service successfully installed.\n");
		CloseServiceHandle(srv);
		CloseServiceHandle(scm);
		return 0;
	}
	else 
	{
		printf("Installation of service failed!\n");
		CloseServiceHandle(scm);
		return 1;
	}
}


//************************************************************************************
// Delete this service
//
//************************************************************************************

int deleteSvc(char *srvName)
{
	SC_HANDLE scm, srv;	
	
	scm = OpenSCManager(
			NULL,					// local machine
			NULL,					// SERVICE_ACTIVE_DATABASE
			SC_MANAGER_CONNECT);

	if (scm == NULL) 
	{
		printf("Error opening SC Manager!  Uninstall of service failed.\n");
		return -1;
	}

	srv = OpenService(scm, srvName, DELETE);
	if(srv == NULL)
	{
		printf("Uninstall of service failed!\n"
			   "Maybe the service wasn't installed or you're not Administrator?\n");
		CloseServiceHandle(scm);
		return 0;
	}

	if(DeleteService(srv))
		printf("Service successfully uninstalled.\n");
	else
		printf("Uninstall of service failed!\n");


	CloseServiceHandle(srv);
	CloseServiceHandle(scm);

	return 1;
}


//************************************************************************************
// Returns the current state of this server
//
//************************************************************************************

DWORD getServiceStatus(char *srvName)
{
	SC_HANDLE hScMgr = OpenSCManager(NULL, NULL, GENERIC_READ);
	if(!hScMgr)
		return -1;

	SC_HANDLE hSvc = OpenService(hScMgr, srvName, SERVICE_QUERY_STATUS);
	if(!hSvc)
	{
		CloseServiceHandle(hScMgr);
		return -1;
	}

	SERVICE_STATUS svcStatus;

	if(!QueryServiceStatus(hSvc, &svcStatus))
	{
		CloseServiceHandle(hSvc);
		CloseServiceHandle(hScMgr);
		return -1;
	}

	//printf("Current State = %d\n", svcStatus.dwCurrentState);

	return svcStatus.dwCurrentState;
}


//************************************************************************************
// Update our state to the SCM
//
//************************************************************************************

BOOL updateOurStatus(DWORD dwState, DWORD dwExitCode, DWORD dwProgress)
{
	SERVICE_STATUS srvStatus;

	srvStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	srvStatus.dwCurrentState = dwState;
	srvStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	srvStatus.dwWin32ExitCode = dwExitCode;
	srvStatus.dwServiceSpecificExitCode = 0;
	srvStatus.dwCheckPoint = dwProgress;
	srvStatus.dwWaitHint = 3000;

	return SetServiceStatus(gSvcHandle, &srvStatus);
}


